# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import copy
from functools import reduce

import numpy as np
import pandas as pd

from ...core import ENTITY_TYPE, CHUNK_TYPE, recursive_tile
from ...serialization.serializables import AnyField
from ...tensor.core import TENSOR_TYPE, TENSOR_CHUNK_TYPE, ChunkData, Chunk
from ...tensor.datasource import tensor as astensor
from ...utils import classproperty, get_dtype
from ..align import (
    align_series_series,
    align_dataframe_series,
    align_dataframe_dataframe,
)
from ..core import (
    DATAFRAME_TYPE,
    SERIES_TYPE,
    DATAFRAME_CHUNK_TYPE,
    SERIES_CHUNK_TYPE,
    is_chunk_meta_lazy,
)
from ..initializer import Series, DataFrame
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..ufunc.tensor import TensorUfuncMixin
from ..utils import (
    parse_index,
    infer_dtypes,
    infer_dtype,
    infer_index_value,
    build_empty_df,
)


class DataFrameBinOpMixin(DataFrameOperandMixin):
    @classmethod
    def _tile_both_dataframes(cls, op):
        # if both of the inputs are DataFrames, axis is just ignored
        left, right = op.lhs, op.rhs
        df = op.outputs[0]

        nsplits, out_shape, left_chunks, right_chunks = align_dataframe_dataframe(
            left, right
        )
        out_chunk_indexes = itertools.product(*(range(s) for s in out_shape))

        out_chunks = []
        for idx, left_chunk, right_chunk in zip(
            out_chunk_indexes, left_chunks, right_chunks
        ):
            out_chunk = (
                op.copy()
                .reset_key()
                .new_chunk(
                    [left_chunk, right_chunk],
                    shape=(nsplits[0][idx[0]], nsplits[1][idx[1]]),
                    index=idx,
                )
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_dataframes(
            op.inputs,
            df.shape,
            nsplits=tuple(tuple(ns) for ns in nsplits),
            chunks=out_chunks,
            dtypes=df.dtypes,
            index_value=df.index_value,
            columns_value=df.columns_value,
        )

    @classmethod
    def _tile_both_series(cls, op):
        left, right = op.lhs, op.rhs
        df = op.outputs[0]

        nsplits, out_shape, left_chunks, right_chunks = align_series_series(left, right)

        out_chunks = []
        for idx, left_chunk, right_chunk in zip(
            range(out_shape[0]), left_chunks, right_chunks
        ):
            out_chunk = (
                op.copy()
                .reset_key()
                .new_chunk(
                    [left_chunk, right_chunk], shape=(nsplits[0][idx],), index=(idx,)
                )
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_seriess(
            op.inputs,
            df.shape,
            nsplits=tuple(tuple(ns) for ns in nsplits),
            chunks=out_chunks,
            dtype=df.dtype,
            index_value=df.index_value,
            name=df.name,
        )

    @classmethod
    def _tile_dataframe_series(cls, op):
        left, right = op.lhs, op.rhs
        df = op.outputs[0]

        nsplits, out_shape, left_chunks, right_chunks = align_dataframe_series(
            left, right, axis=op.axis
        )
        out_chunk_indexes = itertools.product(*(range(s) for s in out_shape))

        out_chunks = []
        for out_idx, df_chunk in zip(out_chunk_indexes, left_chunks):
            if op.axis == "columns" or op.axis == 1:
                series_chunk = right_chunks[out_idx[1]]
                kw = {
                    "shape": (nsplits[0][out_idx[0]], nsplits[1][out_idx[1]]),
                    "index_value": df_chunk.index_value,
                    "dtypes_value": df_chunk.dtypes_value,
                }
            else:
                series_chunk = right_chunks[out_idx[0]]
                kw = {
                    "shape": (nsplits[0][out_idx[0]], nsplits[1][out_idx[1]]),
                    "columns_value": df_chunk.columns_value,
                    "dtypes_value": df_chunk.dtypes_value,
                }
            out_chunk = (
                op.copy()
                .reset_key()
                .new_chunk([df_chunk, series_chunk], index=out_idx, **kw)
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_dataframes(
            op.inputs,
            df.shape,
            nsplits=tuple(tuple(ns) for ns in nsplits),
            chunks=out_chunks,
            dtypes=df.dtypes,
            index_value=df.index_value,
            columns_value=df.columns_value,
        )

    @classmethod
    def _tile_series_dataframe(cls, op):
        left, right = op.lhs, op.rhs
        df = op.outputs[0]

        nsplits, out_shape, right_chunks, left_chunks = align_dataframe_series(
            right, left, axis=op.axis
        )
        out_chunk_indexes = itertools.product(*(range(s) for s in out_shape))

        out_chunks = []
        for out_idx, df_chunk in zip(out_chunk_indexes, right_chunks):
            if op.axis == "columns" or op.axis == 1:
                series_chunk = left_chunks[out_idx[1]]
                kw = {
                    "shape": (df_chunk.shape[0], np.nan),
                    "index_value": df_chunk.index_value,
                    "dtypes_value": df_chunk.dtypes_value,
                }
            else:
                series_chunk = left_chunks[out_idx[0]]
                kw = {
                    "shape": (df_chunk.shape[0], np.nan),
                    "index_value": df_chunk.index_value,
                    "dtypes_value": df_chunk.dtypes_value,
                }
            out_chunk = (
                op.copy()
                .reset_key()
                .new_chunk([series_chunk, df_chunk], index=out_idx, **kw)
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_dataframes(
            op.inputs,
            df.shape,
            nsplits=tuple(tuple(ns) for ns in nsplits),
            chunks=out_chunks,
            dtypes=df.dtypes,
            index_value=df.index_value,
            columns_value=df.columns_value,
        )

    @classmethod
    def _tile_scalar(cls, op):
        tileable = op.rhs if pd.api.types.is_scalar(op.lhs) else op.lhs
        df = op.outputs[0]
        out_chunks = []
        lazy_chunk_meta = is_chunk_meta_lazy(tileable.chunks[0])
        for chunk in tileable.chunks:
            out_op = op.copy().reset_key()
            if chunk.ndim == 2:
                if lazy_chunk_meta:
                    out_chunk = out_op.new_chunk(
                        [chunk],
                        shape=chunk.shape,
                        index=chunk.index,
                    )
                    out_chunk._set_tileable_meta(
                        tileable_key=df.key,
                        nsplits=tileable.nsplits,
                        index_value=df.index_value,
                        columns_value=df.columns_value,
                        dtypes=df.dtypes,
                    )
                else:
                    out_chunk = out_op.new_chunk(
                        [chunk],
                        shape=chunk.shape,
                        index=chunk.index,
                        dtypes=chunk.dtypes,
                        index_value=chunk.index_value,
                        columns_value=getattr(chunk, "columns_value"),
                    )
            else:
                if lazy_chunk_meta:
                    out_chunk = out_op.new_chunk(
                        [chunk],
                        shape=chunk.shape,
                        index=chunk.index,
                        dtype=chunk.dtype,
                        name=getattr(chunk, "name"),
                    )
                    out_chunk._set_tileable_meta(
                        tileable_key=df.key,
                        nsplits=tileable.nsplits,
                        index_value=df.index_value,
                    )
                else:
                    out_chunk = out_op.new_chunk(
                        [chunk],
                        shape=chunk.shape,
                        index=chunk.index,
                        dtype=chunk.dtype,
                        index_value=chunk.index_value,
                        name=getattr(chunk, "name"),
                    )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = df.params.copy()
        params["chunks"] = out_chunks
        params["nsplits"] = tileable.nsplits
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def _tile_with_tensor(cls, op):
        out = op.outputs[0]
        axis = op.axis
        if axis is None:
            axis = 0

        rhs_is_tensor = isinstance(op.rhs, TENSOR_TYPE)
        tensor, other = (op.rhs, op.lhs) if rhs_is_tensor else (op.lhs, op.rhs)
        if tensor.shape == other.shape:
            tensor = yield from recursive_tile(tensor.rechunk(other.nsplits))
        else:
            # shape differs only when dataframe add 1-d tensor, we need rechunk on columns axis.
            if axis in ["columns", 1] and other.ndim == 1:
                # force axis == 0 if it's Series other than DataFrame
                axis = 0
            rechunk_size = (
                other.nsplits[1] if axis == "columns" or axis == 1 else other.nsplits[0]
            )
            if tensor.ndim > 0:
                tensor = yield from recursive_tile(tensor.rechunk((rechunk_size,)))

        out_chunks = []
        for out_index in itertools.product(*(map(range, other.chunk_shape))):
            tensor_chunk = tensor.cix[out_index[: tensor.ndim]]
            other_chunk = other.cix[out_index]
            out_op = op.copy().reset_key()
            inputs = (
                [other_chunk, tensor_chunk]
                if rhs_is_tensor
                else [tensor_chunk, other_chunk]
            )
            if isinstance(other_chunk, DATAFRAME_CHUNK_TYPE):
                cum_splits = [0] + np.cumsum(other.nsplits[1]).tolist()
                start = cum_splits[out_index[1]]
                end = cum_splits[out_index[1] + 1]
                chunk_dtypes = out.dtypes.iloc[start:end]
                out_chunk = out_op.new_chunk(
                    inputs,
                    shape=other_chunk.shape,
                    index=other_chunk.index,
                    dtypes=chunk_dtypes,
                    index_value=other_chunk.index_value,
                    columns_value=other_chunk.columns_value,
                )
            else:
                out_chunk = out_op.new_chunk(
                    inputs,
                    shape=other_chunk.shape,
                    index=other_chunk.index,
                    dtype=out.dtype,
                    index_value=other_chunk.index_value,
                    name=other_chunk.name,
                )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        if isinstance(other, SERIES_TYPE):
            return new_op.new_seriess(
                op.inputs,
                other.shape,
                nsplits=other.nsplits,
                dtype=out.dtype,
                name=other.name,
                index_value=other.index_value,
                chunks=out_chunks,
            )
        else:
            return new_op.new_dataframes(
                op.inputs,
                other.shape,
                nsplits=other.nsplits,
                dtypes=out.dtypes,
                index_value=other.index_value,
                columns_value=other.columns_value,
                chunks=out_chunks,
            )

    @classmethod
    def tile(cls, op):
        if len(op.inputs) < 2:
            return cls._tile_scalar(op)
        elif isinstance(op.inputs[0], DATAFRAME_TYPE) and isinstance(
            op.inputs[1], DATAFRAME_TYPE
        ):
            return cls._tile_both_dataframes(op)
        elif isinstance(op.inputs[0], SERIES_TYPE) and isinstance(
            op.inputs[1], SERIES_TYPE
        ):
            return cls._tile_both_series(op)
        elif isinstance(op.inputs[0], DATAFRAME_TYPE) and isinstance(
            op.inputs[1], SERIES_TYPE
        ):
            return cls._tile_dataframe_series(op)
        elif isinstance(op.inputs[0], SERIES_TYPE) and isinstance(
            op.inputs[1], DATAFRAME_TYPE
        ):
            return cls._tile_series_dataframe(op)
        elif isinstance(op.inputs[0], TENSOR_TYPE) or isinstance(
            op.inputs[1], TENSOR_TYPE
        ):
            return (yield from cls._tile_with_tensor(op))

    @classmethod
    def execute(cls, ctx, op):
        if getattr(cls, "_func_name", None) is not None:
            if len(op.inputs) == 2:
                df, other = ctx[op.inputs[0].key], ctx[op.inputs[1].key]
                if isinstance(op.inputs[0], SERIES_CHUNK_TYPE) and isinstance(
                    op.inputs[1], DATAFRAME_CHUNK_TYPE
                ):
                    df, other = other, df
                    func_name = getattr(cls, "_rfunc_name")
                else:
                    func_name = getattr(cls, "_func_name")
            elif pd.api.types.is_scalar(op.lhs) or isinstance(op.lhs, np.ndarray):
                df = ctx[op.rhs.key]
                other = op.lhs
                func_name = getattr(cls, "_rfunc_name")
            else:
                df = ctx[op.lhs.key]
                other = op.rhs
                func_name = getattr(cls, "_func_name")
            if df.ndim == 2:
                kw = dict(axis=op.axis)
            else:
                kw = dict()
            if op.fill_value is not None:
                # comparison function like eq does not have `fill_value`
                kw["fill_value"] = op.fill_value
            if op.level is not None:
                # logical function like and may don't have `level` (for Series type)
                kw["level"] = op.level
            if hasattr(other, "ndim") and other.ndim == 0:
                other = other.item()
            ctx[op.outputs[0].key] = getattr(df, func_name)(other, **kw)
        else:
            inputs_iter = iter(op.inputs)
            if not pd.api.types.is_scalar(op.lhs):
                lhs = ctx[next(inputs_iter).key]
            else:
                lhs = op.lhs
            if not pd.api.types.is_scalar(op.rhs):
                rhs = ctx[next(inputs_iter).key]
            else:
                rhs = op.rhs
            ctx[op.outputs[0].key] = cls._operator(
                lhs, rhs
            )  # pylint: disable=too-many-function-args

    @classproperty
    def _operator(self):
        raise NotImplementedError

    @classmethod
    def _calc_properties(cls, x1, x2=None, axis="columns"):
        is_chunk = isinstance(x1, CHUNK_TYPE)

        if isinstance(x1, (DATAFRAME_TYPE, DATAFRAME_CHUNK_TYPE)) and (
            x2 is None
            or pd.api.types.is_scalar(x2)
            or isinstance(x2, (TENSOR_TYPE, TENSOR_CHUNK_TYPE))
        ):
            if not is_chunk:
                if pd.api.types.is_scalar(x2):
                    dtypes = cls._operator(build_empty_df(x1.dtypes), x2).dtypes
                elif x1.dtypes is not None and isinstance(x2, TENSOR_TYPE):
                    dtypes = pd.Series(
                        [infer_dtype(dt, x2.dtype, cls._operator) for dt in x1.dtypes],
                        index=x1.dtypes.index,
                    )
                else:  # pragma: no cover
                    dtypes = x1.dtypes
                return {
                    "shape": x1.shape,
                    "dtypes": dtypes,
                    "columns_value": x1.columns_value,
                    "index_value": x1.index_value,
                }
            else:
                return {"shape": x1.shape}

        if isinstance(x1, (SERIES_TYPE, SERIES_CHUNK_TYPE)) and (
            x2 is None
            or pd.api.types.is_scalar(x2)
            or isinstance(x2, (TENSOR_TYPE, TENSOR_CHUNK_TYPE))
        ):
            x2_dtype = x2.dtype if hasattr(x2, "dtype") else type(x2)
            x2_dtype = get_dtype(x2_dtype)
            if hasattr(cls, "return_dtype"):
                dtype = cls.return_dtype
            else:
                dtype = infer_dtype(x1.dtype, x2_dtype, cls._operator)
            ret = {"shape": x1.shape, "dtype": dtype}
            if pd.api.types.is_scalar(x2) or (
                hasattr(x2, "ndim") and (x2.ndim == 0 or x2.ndim == 1)
            ):
                ret["name"] = x1.name
            if not is_chunk:
                ret["index_value"] = x1.index_value
            return ret

        if isinstance(x1, (DATAFRAME_TYPE, DATAFRAME_CHUNK_TYPE)) and isinstance(
            x2, (DATAFRAME_TYPE, DATAFRAME_CHUNK_TYPE)
        ):
            index_shape, column_shape, dtypes, columns, index = (
                np.nan,
                np.nan,
                None,
                None,
                None,
            )

            if (
                x1.columns_value is not None
                and x2.columns_value is not None
                and x1.columns_value.key == x2.columns_value.key
            ):
                dtypes = pd.Series(
                    [
                        infer_dtype(dt1, dt2, cls._operator)
                        for dt1, dt2 in zip(x1.dtypes, x2.dtypes)
                    ],
                    index=x1.dtypes.index,
                )
                columns = copy.copy(x1.columns_value)
                column_shape = len(dtypes)
            elif x1.dtypes is not None and x2.dtypes is not None:
                dtypes = infer_dtypes(x1.dtypes, x2.dtypes, cls._operator)
                columns = parse_index(dtypes.index, store_data=True)
                column_shape = len(dtypes)
            if x1.index_value is not None and x2.index_value is not None:
                if x1.index_value.key == x2.index_value.key:
                    index = copy.copy(x1.index_value)
                    index_shape = x1.shape[0]
                else:
                    index = infer_index_value(x1.index_value, x2.index_value)
                    if index.key == x1.index_value.key == x2.index_value.key and (
                        not np.isnan(x1.shape[0]) or not np.isnan(x2.shape[0])
                    ):
                        index_shape = (
                            x1.shape[0] if not np.isnan(x1.shape[0]) else x2.shape[0]
                        )

            return {
                "shape": (index_shape, column_shape),
                "dtypes": dtypes,
                "columns_value": columns,
                "index_value": index,
            }

        if isinstance(x1, (DATAFRAME_TYPE, DATAFRAME_CHUNK_TYPE)) and isinstance(
            x2, (SERIES_TYPE, SERIES_CHUNK_TYPE)
        ):
            if axis == "columns" or axis == 1:
                index_shape = x1.shape[0]
                index = x1.index_value
                column_shape, dtypes, columns = np.nan, None, None
                if x1.columns_value is not None and x1.index_value is not None:
                    if x1.columns_value.key == x2.index_value.key:
                        dtypes = pd.Series(
                            [
                                infer_dtype(dt, x2.dtype, cls._operator)
                                for dt in x1.dtypes
                            ],
                            index=x1.dtypes.index,
                        )
                        columns = copy.copy(x1.columns_value)
                        column_shape = len(dtypes)
                    else:  # pragma: no cover
                        dtypes = x1.dtypes  # FIXME
                        columns = infer_index_value(x1.columns_value, x2.index_value)
                        column_shape = np.nan
            else:
                assert axis == "index" or axis == 0
                column_shape = x1.shape[1]
                columns = x1.columns_value
                dtypes = x1.dtypes
                index_shape, index = np.nan, None
                if x1.index_value is not None and x1.index_value is not None:
                    if x1.index_value.key == x2.index_value.key:
                        dtypes = pd.Series(
                            [
                                infer_dtype(dt, x2.dtype, cls._operator)
                                for dt in x1.dtypes
                            ],
                            index=x1.dtypes.index,
                        )
                        index = copy.copy(x1.index_value)
                        index_shape = x1.shape[0]
                    else:
                        if x1.dtypes is not None:
                            dtypes = pd.Series(
                                [
                                    infer_dtype(dt, x2.dtype, cls._operator)
                                    for dt in x1.dtypes
                                ],
                                index=x1.dtypes.index,
                            )
                        index = infer_index_value(x1.index_value, x2.index_value)
                        index_shape = np.nan
            return {
                "shape": (index_shape, column_shape),
                "dtypes": dtypes,
                "columns_value": columns,
                "index_value": index,
            }

        if isinstance(x1, (SERIES_TYPE, SERIES_CHUNK_TYPE)) and isinstance(
            x2, (SERIES_TYPE, SERIES_CHUNK_TYPE)
        ):
            index_shape, dtype, index = np.nan, None, None

            dtype = infer_dtype(x1.dtype, x2.dtype, cls._operator)
            if x1.index_value is not None and x2.index_value is not None:
                if x1.index_value.key == x2.index_value.key:
                    index = copy.copy(x1.index_value)
                    index_shape = x1.shape[0]
                else:
                    index = infer_index_value(x1.index_value, x2.index_value)
                    if index.key == x1.index_value.key == x2.index_value.key and (
                        not np.isnan(x1.shape[0]) or not np.isnan(x2.shape[0])
                    ):
                        index_shape = (
                            x1.shape[0] if not np.isnan(x1.shape[0]) else x2.shape[0]
                        )

            ret = {"shape": (index_shape,), "dtype": dtype, "index_value": index}
            if x1.name == x2.name:
                ret["name"] = x1.name
            return ret

        raise NotImplementedError("Unknown combination of parameters")

    def _new_chunks(self, inputs, kws=None, **kw):
        property_inputs = [
            inp
            for inp in inputs
            if isinstance(
                inp, (DATAFRAME_CHUNK_TYPE, SERIES_CHUNK_TYPE, TENSOR_CHUNK_TYPE)
            )
        ]
        # use first two to infer(for tree operand)
        property_inputs = property_inputs[:2]
        if len(property_inputs) == 1:
            properties = self._calc_properties(*property_inputs)
        elif any(inp.ndim == 2 for inp in property_inputs):
            df1, df2 = (
                property_inputs
                if isinstance(property_inputs[0], DATAFRAME_CHUNK_TYPE)
                else reversed(property_inputs)
            )
            properties = self._calc_properties(df1, df2, axis=self.axis)
        else:
            if property_inputs[0].ndim < property_inputs[1].ndim or isinstance(
                property_inputs[0], (TENSOR_TYPE, TENSOR_CHUNK_TYPE)
            ):
                property_inputs = reversed(property_inputs)
            properties = self._calc_properties(*property_inputs)

        inputs = [inp for inp in inputs if isinstance(inp, (Chunk, ChunkData))]

        shape = properties.pop("shape")
        if "shape" in kw:
            shape = kw.pop("shape")

        for prop, value in properties.items():
            if kw.get(prop, None) is None:
                kw[prop] = value

        return super()._new_chunks(inputs, shape=shape, kws=kws, **kw)

    @staticmethod
    def _process_input(x):
        if isinstance(x, (DATAFRAME_TYPE, SERIES_TYPE)) or pd.api.types.is_scalar(x):
            return x
        elif isinstance(x, pd.Series):
            return Series(x)
        elif isinstance(x, pd.DataFrame):
            return DataFrame(x)
        elif isinstance(x, (list, tuple, np.ndarray, TENSOR_TYPE)):
            return astensor(x)
        raise NotImplementedError

    def _check_inputs(self, x1, x2):
        if isinstance(x1, TENSOR_TYPE) or isinstance(x2, TENSOR_TYPE):
            tensor, other = (x1, x2) if isinstance(x1, TENSOR_TYPE) else (x2, x1)
            if isinstance(other, DATAFRAME_TYPE):
                if self.axis == "index" or self.axis == 0:
                    other_shape = tuple(reversed(other.shape))
                else:
                    other_shape = other.shape
                if tensor.ndim == 2 and tensor.shape != other_shape:
                    raise ValueError(
                        f"Unable to coerce to DataFrame, shape must be {other_shape}: "
                        f"given {tensor.shape}"
                    )
                elif tensor.ndim == 1 and tensor.shape[0] != other_shape[1]:
                    raise ValueError(
                        f"Unable to coerce to Series, length must be {other_shape[1]}: "
                        f"given {tensor.shape[0]}"
                    )
                elif tensor.ndim > 2:
                    raise ValueError(
                        "Unable to coerce to Series/DataFrame, dim must be <= 2"
                    )
            if isinstance(other, SERIES_TYPE):
                if tensor.ndim == 1 and (tensor.shape[0] != other.shape[0]):
                    raise ValueError(
                        f"Unable to coerce to Series, length must be {other.shape[0]}: "
                        f"given {tensor.shape[0]}"
                    )
                elif tensor.ndim > 1:
                    raise ValueError("Unable to coerce to Series, dim must be 1")

    def _call(self, x1, x2):
        self._check_inputs(x1, x2)
        if isinstance(x1, DATAFRAME_TYPE) or isinstance(x2, DATAFRAME_TYPE):
            df1, df2 = (x1, x2) if isinstance(x1, DATAFRAME_TYPE) else (x2, x1)
            kw = self._calc_properties(df1, df2, axis=self.axis)
            if not pd.api.types.is_scalar(df2):
                return self.new_dataframe([x1, x2], **kw)
            else:
                return self.new_dataframe([df1], **kw)
        if isinstance(x1, SERIES_TYPE) or isinstance(x2, SERIES_TYPE):
            s1, s2 = (x1, x2) if isinstance(x1, SERIES_TYPE) else (x2, x1)
            kw = self._calc_properties(s1, s2)
            if not pd.api.types.is_scalar(s2):
                return self.new_series([x1, x2], **kw)
            else:
                return self.new_series([s1], **kw)
        raise NotImplementedError(
            "Only support add dataframe, series or scalar for now"
        )

    def __call__(self, x1, x2):
        x1 = self._process_input(x1)
        x2 = self._process_input(x2)
        if isinstance(x1, SERIES_TYPE) and isinstance(x2, DATAFRAME_TYPE):
            # reject invoking series's op on dataframe
            raise NotImplementedError
        return self._call(x1, x2)

    def rcall(self, x1, x2):
        x1 = self._process_input(x1)
        x2 = self._process_input(x2)
        if isinstance(x1, SERIES_TYPE) and isinstance(x2, DATAFRAME_TYPE):
            # reject invoking series's op on dataframe
            raise NotImplementedError
        return self._call(x2, x1)


class DataFrameBinOp(DataFrameOperand, DataFrameBinOpMixin):
    _axis = AnyField("axis")
    _level = AnyField("level")
    _fill_value = AnyField("fill_value")
    _lhs = AnyField("lhs")
    _rhs = AnyField("rhs")

    def __init__(
        self,
        axis=None,
        level=None,
        fill_value=None,
        output_types=None,
        lhs=None,
        rhs=None,
        **kw,
    ):
        super().__init__(
            _axis=axis,
            _level=level,
            _fill_value=fill_value,
            _output_types=output_types,
            _lhs=lhs,
            _rhs=rhs,
            **kw,
        )

    @property
    def axis(self):
        return self._axis

    @property
    def level(self):
        return self._level

    @property
    def fill_value(self):
        return self._fill_value

    @property
    def lhs(self):
        return self._lhs

    @property
    def rhs(self):
        return self._rhs

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if len(self._inputs) == 2:
            self._lhs = self._inputs[0]
            self._rhs = self._inputs[1]
        else:
            if isinstance(self._lhs, ENTITY_TYPE):
                self._lhs = self._inputs[0]
            elif pd.api.types.is_scalar(self._lhs):
                self._rhs = self._inputs[0]


class DataFrameUnaryOpMixin(DataFrameOperandMixin):
    __slots__ = ()

    @classmethod
    def tile(cls, op):
        in_df = op.inputs[0]
        out_df = op.outputs[0]

        out_chunks = []
        index_dtypes_cache = dict()
        for in_chunk in in_df.chunks:
            out_op = op.copy().reset_key()
            if out_df.ndim == 2:
                try:
                    dtypes = index_dtypes_cache[in_chunk.index[1]]
                except KeyError:
                    dtypes = out_df.dtypes[in_chunk.columns_value.to_pandas()]
                    index_dtypes_cache[in_chunk.index[1]] = dtypes

                out_chunk = out_op.new_chunk(
                    [in_chunk],
                    shape=in_chunk.shape,
                    dtypes=dtypes,
                    index=in_chunk.index,
                    index_value=in_chunk.index_value,
                    columns_value=in_chunk.columns_value,
                )
            else:
                out_chunk = out_op.new_chunk(
                    [in_chunk],
                    shape=in_chunk.shape,
                    index=in_chunk.index,
                    dtype=in_chunk.dtype,
                    index_value=in_chunk.index_value,
                    name=in_chunk.name,
                )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        kw = out_df.params
        kw["nsplits"] = in_df.nsplits
        kw["chunks"] = out_chunks
        return new_op.new_tileables(op.inputs, kws=[kw])

    @classmethod
    def execute(cls, ctx, op):
        df = ctx[op.inputs[0].key]
        func_name = getattr(cls, "_func_name")
        if hasattr(df, func_name):
            ctx[op.outputs[0].key] = getattr(df, func_name)()
        else:
            ctx[op.outputs[0].key] = getattr(np, func_name)(df)


class DataFrameUnaryOp(DataFrameOperand, DataFrameUnaryOpMixin):
    def __init__(self, output_types=None, **kw):
        super().__init__(_output_types=output_types, **kw)

    @classmethod
    def _get_output_dtype(cls, df):
        if df.ndim == 2:
            return df.dtypes
        else:
            return df.dtype

    def __call__(self, df):
        self.output_types = df.op.output_types
        if df.ndim == 2:
            return self.new_dataframe(
                [df],
                shape=df.shape,
                dtypes=self._get_output_dtype(df),
                columns_value=df.columns_value,
                index_value=df.index_value,
            )
        else:
            series = df
            return self.new_series(
                [series],
                shape=series.shape,
                name=series.name,
                index_value=series.index_value,
                dtype=self._get_output_dtype(series),
            )


class DataFrameArithmeticTreeMixin:
    @classmethod
    def execute(cls, ctx, op):
        inputs = [ctx[c.key] for c in op.inputs]
        ctx[op.outputs[0].key] = reduce(op._operator, inputs)

    def _set_inputs(self, inputs):
        inputs = self._get_inputs_data(inputs)
        setattr(self, "_inputs", inputs)


class DataFrameUnaryUfunc(DataFrameUnaryOp, TensorUfuncMixin):
    pass


class DataFrameBinopUfunc(DataFrameBinOp, TensorUfuncMixin):
    pass



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
from pandas.api.types import is_integer

from ...core import ENTITY_TYPE
from ...tensor import tensor as astensor
from ...tensor.statistics.percentile import percentile
from ..core import DATAFRAME_TYPE, SERIES_TYPE
from ..initializer import DataFrame, Series
from .cut import cut


def qcut(x, q, labels=None, retbins=False, precision=3, duplicate="raise"):
    """
    Quantile-based discretization function.

    Discretize variable into equal-sized buckets based on rank or based
    on sample quantiles. For example 1000 values for 10 quantiles would
    produce a Categorical object indicating quantile membership for each data point.

    Parameters
    ----------
    x : 1d tensor or Series
    q : int or list-like of float
        Number of quantiles. 10 for deciles, 4 for quartiles, etc. Alternately
        array of quantiles, e.g. [0, .25, .5, .75, 1.] for quartiles.
    labels : array or False, default None
        Used as labels for the resulting bins. Must be of the same length as
        the resulting bins. If False, return only integer indicators of the
        bins. If True, raises an error.
    retbins : bool, optional
        Whether to return the (bins, labels) or not. Can be useful if bins
        is given as a scalar.
    precision : int, optional
        The precision at which to store and display the bins labels.
    duplicates : {default 'raise', 'drop'}, optional
        If bin edges are not unique, raise ValueError or drop non-uniques.

    Returns
    -------
    out : Categorical or Series or tensor of integers if labels is False
        The return type (Categorical or Series) depends on the input: a Series
        of type category if input is a Series else Categorical. Bins are
        represented as categories when categorical data is returned.
    bins : tensor of floats
        Returned only if `retbins` is True.

    Notes
    -----
    Out of bounds values will be NA in the resulting Categorical object

    Examples
    --------
    >>> import mars.dataframe as md
    >>> md.qcut(range(5), 4).execute()
    ... # doctest: +ELLIPSIS
    [(-0.001, 1.0], (-0.001, 1.0], (1.0, 2.0], (2.0, 3.0], (3.0, 4.0]]
    Categories (4, interval[float64]): [(-0.001, 1.0] < (1.0, 2.0] ...

    >>> md.qcut(range(5), 3, labels=["good", "medium", "bad"]).execute()
    ... # doctest: +SKIP
    [good, good, medium, bad, bad]
    Categories (3, object): [good < medium < bad]

    >>> md.qcut(range(5), 4, labels=False).execute()
    array([0, 0, 1, 2, 3])
    """
    if is_integer(q):
        q = np.linspace(0, 1, q + 1)

    if isinstance(x, (DATAFRAME_TYPE, SERIES_TYPE, pd.DataFrame, pd.Series)):
        x = DataFrame(x) if x.ndim == 2 else Series(x)
        bins = x.quantile(q)
    else:
        x = astensor(x)
        if isinstance(q, ENTITY_TYPE):
            q = q * 100
        else:
            q = [iq * 100 for iq in q]
        bins = percentile(x, q)

    return cut(
        x,
        bins,
        labels=labels,
        retbins=retbins,
        precision=precision,
        include_lowest=True,
        duplicates=duplicate,
    )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...serialization.serializables import BoolField, FieldTypes, TupleField
from ...utils import tokenize
from .core import EntityData, Entity


class ChunkData(EntityData):
    __slots__ = ()

    is_broadcaster = BoolField("is_broadcaster", default=False)
    # optional fields
    _index = TupleField("index", FieldTypes.uint32)

    def __repr__(self):
        if self.op.stage is None:
            return f"Chunk <op={type(self.op).__name__}, key={self.key}>"
        else:
            return (
                f"Chunk <op={type(self.op).__name__}, "
                f"stage={self.op.stage.name}, key={self.key}>"
            )

    @property
    def index(self):
        return getattr(self, "_index", None)

    @property
    def device(self):
        return self.op.device

    def _update_key(self):
        object.__setattr__(
            self,
            "_key",
            tokenize(
                type(self).__name__,
                *(getattr(self, k, None) for k in self._keys_ if k != "_index"),
            ),
        )


class Chunk(Entity):
    _allow_data_type_ = (ChunkData,)


CHUNK_TYPE = (Chunk, ChunkData)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
from typing import Dict, List, Tuple, Type, Any, Optional

from .communication import get_client_type, Client


class Router:
    """
    Router provides mapping from external address to internal address.
    """

    __slots__ = "_curr_external_addresses", "_local_mapping", "_mapping", "_cache_local"

    _instance: "Router" = None

    @staticmethod
    def set_instance(router: Optional["Router"]):
        # Default router is set when an actor pool started
        Router._instance = router

    @staticmethod
    def get_instance() -> "Router":
        return Router._instance

    @staticmethod
    def get_instance_or_empty() -> "Router":
        return Router._instance or Router(list(), None)

    def __init__(
        self,
        external_addresses: List[str],
        local_address: Optional[str],
        mapping: Dict[str, str] = None,
    ):
        self._curr_external_addresses = external_addresses
        self._local_mapping = dict()
        for addr in self._curr_external_addresses:
            self._local_mapping[addr] = local_address
        if mapping is None:
            mapping = dict()
        self._mapping = mapping
        self._cache_local = threading.local()

    @property
    def _cache(self) -> Dict[Tuple[str, Any], Client]:
        try:
            return self._cache_local.cache
        except AttributeError:
            cache = self._cache_local.cache = dict()
            return cache

    def set_mapping(self, mapping: Dict[str, str]):
        self._mapping = mapping
        self._cache_local = threading.local()

    def add_router(self, router: "Router"):
        self._curr_external_addresses.extend(router._curr_external_addresses)
        self._local_mapping.update(router._local_mapping)
        self._mapping.update(router._mapping)
        self._cache_local = threading.local()

    def remove_router(self, router: "Router"):
        for external_address in router._curr_external_addresses:
            try:
                self._curr_external_addresses.remove(external_address)
            except ValueError:
                pass
        for addr in router._local_mapping:
            self._local_mapping.pop(addr, None)
        for addr in router._mapping:
            self._mapping.pop(addr, None)
        self._cache_local = threading.local()

    @property
    def external_address(self):
        if self._curr_external_addresses:
            return self._curr_external_addresses[0]

    def get_internal_address(self, external_address: str) -> str:
        if external_address in self._curr_external_addresses:
            # local address, use dummy address
            return self._local_mapping.get(external_address)
        # try to lookup inner address from address mapping
        return self._mapping.get(external_address)

    async def get_client(
        self, external_address: str, from_who: Any = None, cached: bool = True, **kw
    ) -> Client:
        if cached and (external_address, from_who) in self._cache:
            cached_client = self._cache[external_address, from_who]
            if cached_client.closed:
                # closed before, ignore it
                del self._cache[external_address, from_who]
            else:
                return cached_client

        address = self.get_internal_address(external_address)
        if address is None:
            # no inner address, just use external address
            address = external_address
        client_type: Type[Client] = get_client_type(address)
        local_address = (
            self._curr_external_addresses[0] if self._curr_external_addresses else None
        )
        client = await client_type.connect(address, local_address=local_address, **kw)
        if cached:
            self._cache[external_address, from_who] = client
        return client



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
import warnings

import requests

from ...session import new_session
from .config import (
    MarsSchedulerSpecConfig,
    MarsWorkerSpecConfig,
    MarsWebSpecConfig,
    MarsJobConfig,
)

try:
    from kubernetes.client.rest import ApiException as K8SApiException
except ImportError:  # pragma: no cover
    K8SApiException = None

KUBEDL_API_VERSION = "kubedl.io/v1alpha1"
KUBEDL_MARS_PLURAL = "marsjobs"


logger = logging.getLogger(__name__)


class KubeDLClusterClient:
    def __init__(self, cluster):
        self._cluster = cluster
        self._endpoint = None
        self._session = None

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def namespace(self):
        return self._cluster.namespace

    @property
    def session(self):
        return self._session

    def start(self):
        self._endpoint = self._cluster.start()
        self._session = new_session(self._endpoint, verify_ssl=self._cluster.verify_ssl)

    def stop(self, wait=False, timeout=0):
        self._cluster.stop(wait=wait, timeout=timeout)


class KubeDLCluster:
    def __init__(
        self,
        kube_api_client=None,
        image=None,
        job_name=None,
        namespace=None,
        scheduler_num=1,
        scheduler_cpu=None,
        scheduler_mem=None,
        worker_num=1,
        worker_cpu=None,
        worker_mem=None,
        worker_spill_paths=None,
        worker_cache_mem=None,
        min_worker_num=None,
        web_num=1,
        web_cpu=None,
        web_mem=None,
        slb_endpoint=None,
        verify_ssl=True,
        timeout=None,
        **kwargs,
    ):
        from kubernetes import client as kube_client

        self._kube_api_client = kube_api_client
        self._custom_api = kube_client.CustomObjectsApi(kube_api_client)

        self._slb_endpoint = slb_endpoint.rstrip("/")
        self._verify_ssl = verify_ssl

        self._job_name = job_name
        self._mars_endpoint = None
        self._namespace = namespace or "default"
        self._image = image
        self._timeout = timeout
        self._extra_volumes = kwargs.pop("extra_volumes", ())
        self._pre_stop_command = kwargs.pop("pre_stop_command", None)
        self._log_when_fail = kwargs.pop("log_when_fail", False)
        self._node_selectors = kwargs.pop("node_selectors", None)

        extra_modules = kwargs.pop("extra_modules", None) or []
        extra_modules = (
            extra_modules.split(",")
            if isinstance(extra_modules, str)
            else extra_modules
        )
        extra_envs = kwargs.pop("extra_env", None) or dict()

        if not verify_ssl:
            extra_envs["KUBE_VERIFY_SSL"] = "0"

        def _override_modules(updates):
            modules = set(extra_modules)
            updates = updates.split(",") if isinstance(updates, str) else updates
            modules.update(updates)
            return sorted(modules)

        def _override_envs(updates):
            ret = extra_envs.copy()
            ret.update(updates)
            return ret

        self._scheduler_num = scheduler_num
        self._scheduler_cpu = scheduler_cpu
        self._scheduler_mem = scheduler_mem
        self._scheduler_extra_modules = _override_modules(
            kwargs.pop("scheduler_extra_modules", [])
        )
        self._scheduler_extra_env = _override_envs(
            kwargs.pop("scheduler_extra_env", None) or dict()
        )

        self._worker_num = worker_num
        self._worker_cpu = worker_cpu
        self._worker_mem = worker_mem
        self._worker_spill_paths = worker_spill_paths
        self._worker_cache_mem = worker_cache_mem
        self._min_worker_num = min_worker_num or worker_num
        self._worker_extra_modules = _override_modules(
            kwargs.pop("worker_extra_modules", [])
        )
        self._worker_extra_env = _override_envs(
            kwargs.pop("worker_extra_env", None) or dict()
        )

        self._web_num = web_num
        self._web_cpu = web_cpu
        self._web_mem = web_mem
        self._web_extra_modules = _override_modules(kwargs.pop("web_extra_modules", []))
        self._web_extra_env = _override_envs(
            kwargs.pop("web_extra_env", None) or dict()
        )

    @property
    def verify_ssl(self):
        return self._verify_ssl

    def _check_if_exist(self):
        if self._job_name is None:
            return False
        try:
            api, version = KUBEDL_API_VERSION.rsplit("/", 1)
            service_obj = self._custom_api.get_namespaced_custom_object_status(
                api, version, self._namespace, KUBEDL_MARS_PLURAL, self._job_name
            )
            if len(service_obj.get("status", dict()).get("conditions", [])) > 0:
                status = service_obj["status"]["conditions"][-1]["type"]
                if status == "Running" or status == "Created":
                    logger.warning(f"Reusing cluster: {self._job_name}")
                    return True
                else:
                    return False
            else:
                return False
        except K8SApiException:
            return False

    def _create_service(self):
        scheduler_cfg = MarsSchedulerSpecConfig(
            self._image,
            self._scheduler_num,
            cpu=self._scheduler_cpu,
            memory=self._scheduler_mem,
            node_selectors=self._node_selectors,
            modules=self._scheduler_extra_modules,
        )
        scheduler_cfg.add_simple_envs(self._scheduler_extra_env)

        worker_cfg = MarsWorkerSpecConfig(
            self._image,
            self._worker_num,
            cpu=self._worker_cpu,
            memory=self._worker_mem,
            cache_mem=self._worker_cache_mem,
            spill_dirs=self._worker_spill_paths,
            node_selectors=self._node_selectors,
            modules=self._worker_extra_modules,
        )
        worker_cfg.add_simple_envs(self._worker_extra_env)

        web_cfg = MarsWebSpecConfig(
            self._image,
            self._web_num,
            cpu=self._web_cpu,
            memory=self._web_mem,
            node_selectors=self._node_selectors,
            modules=self._web_extra_modules,
        )
        web_cfg.add_simple_envs(self._web_extra_env)

        job_cfg = MarsJobConfig(
            job_name=self._job_name,
            scheduler_config=scheduler_cfg,
            worker_config=worker_cfg,
            web_config=web_cfg,
            web_host=self._slb_endpoint,
        )

        api, version = KUBEDL_API_VERSION.rsplit("/", 1)

        cfg_json = job_cfg.build()
        cfg_json["apiVersion"] = KUBEDL_API_VERSION

        response = self._custom_api.create_namespaced_custom_object(
            api, version, self._namespace, KUBEDL_MARS_PLURAL, cfg_json
        )
        self._job_name = response["metadata"]["name"]

    def _wait_service_ready(self):
        self._mars_endpoint = (
            f"{self._slb_endpoint}/mars/{self._namespace}/{self._job_name}-webservice-0"
        )
        logger.warning(f"Kubedl job name: {self._job_name}")
        check_start_time = time.time()
        worker_count_url = self._mars_endpoint + "/api/worker?action=count"
        while True:
            try:
                if self._timeout and time.time() - check_start_time > self._timeout:
                    raise TimeoutError("Check Mars service start timeout")

                if not self._verify_ssl:
                    try:
                        import urllib3

                        urllib3.disable_warnings(
                            urllib3.exceptions.InsecureRequestWarning
                        )
                    except ImportError:  # pragma: no cover
                        pass

                api, version = KUBEDL_API_VERSION.rsplit("/", 1)
                service_obj = self._custom_api.get_namespaced_custom_object_status(
                    api, version, self._namespace, KUBEDL_MARS_PLURAL, self._job_name
                )
                if len(service_obj.get("status", dict()).get("conditions", [])) > 0:
                    if service_obj["status"]["conditions"][-1]["type"] == "Failed":
                        raise SystemError(
                            service_obj["status"]["conditions"][-1]["message"]
                        )

                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        "ignore", message="Unverified HTTPS request"
                    )
                    resp = requests.get(
                        worker_count_url, timeout=1, verify=self._verify_ssl
                    )

                if int(resp.text) >= self._min_worker_num:
                    logger.warning(f"Web endpoint started at {self._mars_endpoint}")
                    break
            except (requests.Timeout, ValueError) as ex:
                if not isinstance(ex, requests.Timeout):
                    time.sleep(0.1)
                pass

    def start(self):
        try:
            if not self._check_if_exist():
                self._create_service()
            self._wait_service_ready()
            return self._mars_endpoint
        except:  # noqa: E722
            self.stop()
            raise

    def stop(self, wait=False, timeout=0):
        from kubernetes import client as kube_client

        custom_api = kube_client.CustomObjectsApi(self._kube_api_client)
        api, version = KUBEDL_API_VERSION.rsplit("/", 1)
        custom_api.delete_namespaced_custom_object(
            api, version, self._namespace, KUBEDL_MARS_PLURAL, self._job_name
        )

        if wait:
            start_time = time.time()
            while True:
                try:
                    custom_api.get_namespaced_custom_object(
                        api,
                        version,
                        self._namespace,
                        KUBEDL_MARS_PLURAL,
                        self._job_name,
                    )
                except K8SApiException as ex:
                    if ex.status != 404:  # pragma: no cover
                        raise
                    break
                else:
                    time.sleep(1)
                    if (
                        timeout and time.time() - start_time > timeout
                    ):  # pragma: no cover
                        raise TimeoutError("Check Mars service stop timeout")


def new_cluster(
    kube_api_client=None,
    image=None,
    scheduler_num=1,
    scheduler_cpu=2,
    scheduler_mem=4 * 1024**3,
    worker_num=1,
    worker_cpu=8,
    worker_mem=32 * 1024**3,
    worker_spill_paths=None,
    worker_cache_mem="45%",
    min_worker_num=None,
    web_num=1,
    web_cpu=1,
    web_mem=4 * 1024**3,
    slb_endpoint=None,
    verify_ssl=True,
    job_name=None,
    timeout=None,
    **kwargs,
):
    worker_spill_paths = worker_spill_paths or ["/tmp/spill-dir"]
    cluster = KubeDLCluster(
        kube_api_client,
        image=image,
        scheduler_num=scheduler_num,
        scheduler_cpu=scheduler_cpu,
        scheduler_mem=scheduler_mem,
        worker_num=worker_num,
        worker_cpu=worker_cpu,
        worker_mem=worker_mem,
        worker_spill_paths=worker_spill_paths,
        worker_cache_mem=worker_cache_mem,
        min_worker_num=min_worker_num,
        web_num=web_num,
        web_cpu=web_cpu,
        web_mem=web_mem,
        slb_endpoint=slb_endpoint,
        verify_ssl=verify_ssl,
        job_name=job_name,
        timeout=timeout,
        **kwargs,
    )
    client = KubeDLClusterClient(cluster)
    client.start()
    return client



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noinspection PyUnresolvedReferences
from ..typing import ChunkType, TileableType, EntityType, OperandType
from .base import ExecutionError
from .entity import (
    Entity,
    EntityData,
    ENTITY_TYPE,
    Chunk,
    ChunkData,
    CHUNK_TYPE,
    Tileable,
    TileableData,
    TILEABLE_TYPE,
    Object,
    ObjectData,
    ObjectChunk,
    ObjectChunkData,
    OBJECT_TYPE,
    OBJECT_CHUNK_TYPE,
    FuseChunk,
    FuseChunkData,
    FUSE_CHUNK_TYPE,
    OutputType,
    register_output_types,
    get_output_types,
    register_fetch_class,
    get_fetch_class,
    get_tileable_types,
    get_chunk_types,
    HasShapeTileable,
    HasShapeTileableData,
    ExecutableTuple,
    _ExecuteAndFetchMixin,
    NotSupportTile,
    register,
    unregister,
    tile,
    recursive_tile,
)

# noinspection PyUnresolvedReferences
from .graph import (
    DirectedGraph,
    DAG,
    GraphContainsCycleError,
    TileableGraph,
    ChunkGraph,
    TileableGraphBuilder,
    ChunkGraphBuilder,
    TileContext,
    TileStatus,
)
from .mode import enter_mode, is_build_mode, is_eager_mode, is_kernel_mode



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes
from ...core import OutputType
from ...serialization.serializables import AnyField, BoolField
from ..core import DATAFRAME_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import validate_axis, parse_index, build_df, build_series


class DataFrameRenameAxis(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.RENAME_AXIS

    _index = AnyField("index")
    _columns = AnyField("columns")
    _copy_value = BoolField("copy_value")
    _level = AnyField("level")

    def __init__(self, index=None, columns=None, copy_value=None, level=None, **kw):
        super().__init__(
            _index=index, _columns=columns, _copy_value=copy_value, _level=level, **kw
        )

    @property
    def index(self):
        return self._index

    @property
    def columns(self):
        return self._columns

    @property
    def copy_value(self):
        return self._copy_value

    @property
    def level(self):
        return self._level

    @staticmethod
    def _update_params(params, obj, mapper, axis, level):
        if obj.ndim == 2:
            test_obj = build_df(obj)
        else:
            test_obj = build_series(obj)

        if level is None:
            test_obj = test_obj.rename_axis(mapper, axis=axis)
        else:
            test_obj.axes[axis].set_names(mapper, level=level, inplace=True)

        if axis == 0:
            params["index_value"] = parse_index(test_obj.index, store_data=False)
        else:
            params["dtypes"] = test_obj.dtypes
            params["columns_value"] = parse_index(test_obj.columns, store_data=True)

    def __call__(self, df_or_series):
        params = df_or_series.params

        if isinstance(df_or_series, DATAFRAME_TYPE):
            self._output_types = [OutputType.dataframe]
        else:
            self._output_types = [OutputType.series]

        if self.index is not None:
            self._update_params(
                params, df_or_series, self.index, axis=0, level=self.level
            )
        else:
            self._update_params(
                params, df_or_series, self.columns, axis=1, level=self.level
            )

        return self.new_tileable([df_or_series], **params)

    @classmethod
    def tile(cls, op: "DataFrameRenameAxis"):
        in_obj = op.inputs[0]
        out_obj = op.outputs[0]

        chunks = []
        idx_cache = dict()
        for c in in_obj.chunks:
            params = c.params
            if op.index is not None:
                try:
                    params["index_value"] = idx_cache[c.index[0]]
                except KeyError:
                    cls._update_params(params, c, op.index, axis=0, level=op.level)
                    idx_cache[c.index[0]] = params["index_value"]
            else:
                try:
                    params["columns_value"], params["dtypes"] = idx_cache[c.index[1]]
                except KeyError:
                    cls._update_params(params, c, op.columns, axis=1, level=op.level)
                    idx_cache[c.index[1]] = params["columns_value"], params["dtypes"]

            new_op = op.copy().reset_key()
            chunks.append(new_op.new_chunk([c], **params))

        new_op = op.copy().reset_key()
        return new_op.new_tileables(
            [in_obj], chunks=chunks, nsplits=in_obj.nsplits, **out_obj.params
        )

    @classmethod
    def execute(cls, ctx, op: "DataFrameRenameAxis"):
        in_data = ctx[op.inputs[0].key]
        if op.index is not None:
            val, axis = op.index, 0
        else:
            val, axis = op.columns, 1

        if op.level is None:
            ctx[op.outputs[0].key] = in_data.rename_axis(
                val, axis=axis, copy=op.copy_value
            )
        else:
            ret = in_data.copy() if op.copy_value else in_data
            ret.axes[axis].set_names(val, level=op.level, inplace=True)
            ctx[op.outputs[0].key] = ret


def rename_axis_with_level(
    df_or_series,
    mapper=None,
    index=None,
    columns=None,
    axis=0,
    copy=True,
    level=None,
    inplace=False,
):
    axis = validate_axis(axis, df_or_series)
    if mapper is not None:
        if axis == 0:
            index = mapper
        else:
            columns = mapper
    op = DataFrameRenameAxis(index=index, columns=columns, copy_value=copy, level=level)
    result = op(df_or_series)
    if not inplace:
        return result
    else:
        df_or_series.data = result.data


def rename_axis(
    df_or_series,
    mapper=None,
    index=None,
    columns=None,
    axis=0,
    copy=True,
    inplace=False,
):
    """
    Set the name of the axis for the index or columns.

    Parameters
    ----------
    mapper : scalar, list-like, optional
        Value to set the axis name attribute.
    index, columns : scalar, list-like, dict-like or function, optional
        A scalar, list-like, dict-like or functions transformations to
        apply to that axis' values.
        Note that the ``columns`` parameter is not allowed if the
        object is a Series. This parameter only apply for DataFrame
        type objects.

        Use either ``mapper`` and ``axis`` to
        specify the axis to target with ``mapper``, or ``index``
        and/or ``columns``.

    axis : {0 or 'index', 1 or 'columns'}, default 0
        The axis to rename.
    copy : bool, default True
        Also copy underlying data.
    inplace : bool, default False
        Modifies the object directly, instead of creating a new Series
        or DataFrame.

    Returns
    -------
    Series, DataFrame, or None
        The same type as the caller or None if `inplace` is True.

    See Also
    --------
    Series.rename : Alter Series index labels or name.
    DataFrame.rename : Alter DataFrame index labels or name.
    Index.rename : Set new names on index.

    Notes
    -----
    ``DataFrame.rename_axis`` supports two calling conventions

    * ``(index=index_mapper, columns=columns_mapper, ...)``
    * ``(mapper, axis={'index', 'columns'}, ...)``

    The first calling convention will only modify the names of
    the index and/or the names of the Index object that is the columns.
    In this case, the parameter ``copy`` is ignored.

    The second calling convention will modify the names of the
    the corresponding index if mapper is a list or a scalar.
    However, if mapper is dict-like or a function, it will use the
    deprecated behavior of modifying the axis *labels*.

    We *highly* recommend using keyword arguments to clarify your
    intent.

    Examples
    --------
    **Series**

    >>> import mars.dataframe as md
    >>> s = md.Series(["dog", "cat", "monkey"])
    >>> s.execute()
    0       dog
    1       cat
    2    monkey
    dtype: object
    >>> s.rename_axis("animal").execute()
    animal
    0    dog
    1    cat
    2    monkey
    dtype: object

    **DataFrame**

    >>> df = md.DataFrame({"num_legs": [4, 4, 2],
    ...                    "num_arms": [0, 0, 2]},
    ...                   ["dog", "cat", "monkey"])
    >>> df.execute()
            num_legs  num_arms
    dog            4         0
    cat            4         0
    monkey         2         2
    >>> df = df.rename_axis("animal")
    >>> df.execute()
            num_legs  num_arms
    animal
    dog            4         0
    cat            4         0
    monkey         2         2
    >>> df = df.rename_axis("limbs", axis="columns")
    >>> df.execute()
    limbs   num_legs  num_arms
    animal
    dog            4         0
    cat            4         0
    monkey         2         2
    """
    return rename_axis_with_level(
        df_or_series,
        mapper=mapper,
        index=index,
        columns=columns,
        axis=axis,
        copy=copy,
        inplace=inplace,
    )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import operator
from functools import reduce

from ....utils import lazy_import
from .mldataset import _rechunk_if_needed
from typing import Dict, List

ray = lazy_import("ray")
# Ray Datasets is available in early preview at ray.data with Ray 1.6+
# (and ray.experimental.data in Ray 1.5)
ray_dataset = lazy_import("ray.data", rename="ray_dataset")


def to_ray_dataset(df, num_shards: int = None):
    """Create a Ray Dataset from Mars DataFrame

    Args:
        df (mars.dataframe.Dataframe): the Mars DataFrame
        num_shards (int, optional): the number of shards that will be created
            for the Ray Dataset. Defaults to None.
            If num_shards is None, chunks will be grouped by nodes where they lie.
            Otherwise, chunks will be grouped by their order in DataFrame.

    Returns:
        a Ray Dataset
    """
    df = _rechunk_if_needed(df, num_shards)
    # chunk_addr_refs is fetched directly rather than in batches
    # during `fetch` procedure, it'll be checked that df has been executed
    # items in chunk_addr_refs are ordered by positions in df
    # while adjacent chunks may belong to different addrs, i.e.
    #       chunk1 for addr1,
    #       chunk2 & chunk3 for addr2,
    #       chunk4 for addr1
    chunk_refs: List["ray.ObjectRef"] = get_chunk_refs(df)
    dataset = ray_dataset.from_pandas_refs(chunk_refs)
    # Hold mars dataframe to avoid mars dataframe and ray object gc.
    dataset.dataframe = df

    def __getstate__():
        state = dataset.__dict__.copy()
        state.pop("dataframe", None)
        return state

    # `dataframe` is not serializable by ray.
    dataset.__getstate__ = __getstate__
    return dataset


def get_chunk_refs(df):
    fetched_infos: Dict[str, List] = df.fetch_infos(["object_refs"])
    object_refs = reduce(operator.concat, fetched_infos["object_refs"])
    return object_refs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...serialization.serializables import KeyField, Int32Field
from ..array_utils import as_same_device, device
from ..core import TensorOrder
from ..datasource import tensor as astensor
from ..operands import TensorHasInput, TensorOperandMixin
from ..utils import validate_axis, reverse_order


def _swap(it, axis1, axis2):
    new_it = list(it)
    new_it[axis1], new_it[axis2] = it[axis2], it[axis1]

    return tuple(new_it)


class TensorSwapAxes(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.SWAPAXES

    _input = KeyField("input")
    _axis1 = Int32Field("axis1")
    _axis2 = Int32Field("axis2")

    def __init__(self, axis1=None, axis2=None, **kw):
        super().__init__(_axis1=axis1, _axis2=axis2, create_view=True, **kw)

    @property
    def axis1(self):
        return self._axis1

    @property
    def axis2(self):
        return self._axis2

    def __call__(self, a):
        axis1, axis2 = self._axis1, self._axis2
        if (axis1 == 0 and axis2 == a.ndim - 1) or (axis1 == a.ndim - 1 and axis2 == 0):
            tensor_order = reverse_order(a.order)
        else:
            tensor_order = TensorOrder.C_ORDER
        shape = _swap(a.shape, self.axis1, self.axis2)
        return self.new_tensor([a], shape, order=tensor_order)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def on_output_modify(self, new_output):
        op = TensorSwapAxes(
            axis1=self._axis2,
            axis2=self._axis1,
            dtype=new_output.dtype,
            sparse=new_output.issparse(),
        )
        return op(new_output)

    def on_input_modify(self, new_input):
        op = self.copy().reset_key()
        return op(new_input)

    @classmethod
    def tile(cls, op):
        axis1, axis2 = op.axis1, op.axis2
        in_tensor = op.inputs[0]
        out_tensor = op.outputs[0]

        out_chunks = []
        for c in in_tensor.chunks:
            chunk_shape = _swap(c.shape, axis1, axis2)
            chunk_idx = _swap(c.index, axis1, axis2)
            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk(
                [c], shape=chunk_shape, index=chunk_idx, order=out_tensor.order
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        nsplits = _swap(in_tensor.nsplits, axis1, axis2)
        return new_op.new_tensors(
            [in_tensor],
            out_tensor.shape,
            order=out_tensor.order,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def execute(cls, ctx, op):
        (x,), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        axis1, axis2 = op.axis1, op.axis2
        with device(device_id):
            ctx[op.outputs[0].key] = xp.swapaxes(x, axis1, axis2)


def swapaxes(a, axis1, axis2):
    """
    Interchange two axes of a tensor.

    Parameters
    ----------
    a : array_like
        Input tensor.
    axis1 : int
        First axis.
    axis2 : int
        Second axis.

    Returns
    -------
    a_swapped : Tensor
        If `a` is a Tensor, then a view of `a` is
        returned; otherwise a new tensor is created.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.array([[1,2,3]])
    >>> mt.swapaxes(x,0,1).execute()
    array([[1],
           [2],
           [3]])

    >>> x = mt.array([[[0,1],[2,3]],[[4,5],[6,7]]])
    >>> x.execute()
    array([[[0, 1],
            [2, 3]],
           [[4, 5],
            [6, 7]]])

    >>> mt.swapaxes(x,0,2).execute()
    array([[[0, 4],
            [2, 6]],
           [[1, 5],
            [3, 7]]])

    """
    a = astensor(a)
    axis1 = validate_axis(a.ndim, axis1)
    axis2 = validate_axis(a.ndim, axis2)

    if axis1 == axis2:
        return a

    op = TensorSwapAxes(axis1, axis2, dtype=a.dtype, sparse=a.issparse())
    return op(a)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .sort import _validate_sort_arguments, TensorSort


def argsort(a, axis=-1, kind=None, parallel_kind=None, psrs_kinds=None, order=None):
    """
    Returns the indices that would sort a tensor.

    Perform an indirect sort along the given axis using the algorithm specified
    by the `kind` keyword. It returns a tensor of indices of the same shape as
    `a` that index data along the given axis in sorted order.

    Parameters
    ----------
    a : array_like
        Tensor to sort.
    axis : int or None, optional
        Axis along which to sort.  The default is -1 (the last axis). If None,
        the flattened tensor is used.
    kind : {'quicksort', 'mergesort', 'heapsort', 'stable'}, optional
        Sorting algorithm. The default is 'quicksort'. Note that both 'stable'
        and 'mergesort' use timsort under the covers and, in general, the
        actual implementation will vary with data type. The 'mergesort' option
        is retained for backwards compatibility.

        .. versionchanged:: 1.15.0.
           The 'stable' option was added.
    order : str or list of str, optional
        When `a` is a tensor with fields defined, this argument specifies
        which fields to compare first, second, etc.  A single field can
        be specified as a string, and not all fields need be specified,
        but unspecified fields will still be used, in the order in which
        they come up in the dtype, to break ties.

    Returns
    -------
    index_tensor : Tensor, int
        Tensor of indices that sort `a` along the specified `axis`.
        If `a` is one-dimensional, ``a[index_tensor]`` yields a sorted `a`.
        More generally, ``np.take_along_axis(a, index_tensor, axis=axis)``
        always yields the sorted `a`, irrespective of dimensionality.

    See Also
    --------
    sort : Describes sorting algorithms used.
    lexsort : Indirect stable sort with multiple keys.
    Tensor.sort : Inplace sort.
    argpartition : Indirect partial sort.

    Notes
    -----
    See `sort` for notes on the different sorting algorithms.

    Examples
    --------
    One dimensional tensor:

    >>> import mars.tensor as mt
    >>> x = mt.array([3, 1, 2])
    >>> mt.argsort(x).execute()
    array([1, 2, 0])

    Two-dimensional tensor:

    >>> x = mt.array([[0, 3], [2, 2]])
    >>> x.execute()
    array([[0, 3],
           [2, 2]])

    >>> ind = mt.argsort(x, axis=0)  # sorts along first axis (down)
    >>> ind.execute()
    array([[0, 1],
           [1, 0]])
    #>>> mt.take_along_axis(x, ind, axis=0).execute()  # same as np.sort(x, axis=0)
    #array([[0, 2],
    #       [2, 3]])

    >>> ind = mt.argsort(x, axis=1)  # sorts along last axis (across)
    >>> ind.execute()
    array([[0, 1],
           [0, 1]])
    #>>> mt.take_along_axis(x, ind, axis=1).execute()  # same as np.sort(x, axis=1)
    #array([[0, 3],
    #       [2, 2]])

    Indices of the sorted elements of a N-dimensional array:

    >>> ind = mt.unravel_index(mt.argsort(x, axis=None), x.shape)
    >>> ind.execute9)
    (array([0, 1, 1, 0]), array([0, 0, 1, 1]))
    >>> x[ind].execute()  # same as np.sort(x, axis=None)
    array([0, 2, 2, 3])

    Sorting with keys:

    >>> x = mt.array([(1, 0), (0, 1)], dtype=[('x', '<i4'), ('y', '<i4')])
    >>> x.execute()
    array([(1, 0), (0, 1)],
          dtype=[('x', '<i4'), ('y', '<i4')])

    >>> mt.argsort(x, order=('x','y')).execute()
    array([1, 0])

    >>> mt.argsort(x, order=('y','x')).execute()
    array([0, 1])

    """
    a, axis, kind, parallel_kind, psrs_kinds, order = _validate_sort_arguments(
        a, axis, kind, parallel_kind, psrs_kinds, order
    )

    op = TensorSort(
        axis=axis,
        kind=kind,
        parallel_kind=parallel_kind,
        order=order,
        psrs_kinds=psrs_kinds,
        return_value=False,
        return_indices=True,
        dtype=a.dtype,
        gpu=a.op.gpu,
    )
    return op(a)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...utils import classproperty
from .core import DataFrameBinopUfunc
from .docstring import bin_compare_doc


class DataFrameLessEqual(DataFrameBinopUfunc):
    _op_type_ = OperandDef.LE

    _func_name = "le"
    _rfunc_name = "ge"

    return_dtype = np.dtype(bool)

    @classproperty
    def _operator(self):
        return lambda lhs, rhs: lhs.le(rhs)

    @classproperty
    def tensor_op_type(self):
        from ...tensor.arithmetic import TensorLessEqual

        return TensorLessEqual


_le_example = """
>>> a.le(b, fill_value=0).execute()
a    False
b     True
c     True
d    False
e    False
f     True
dtype: bool
"""


@bin_compare_doc("Less than or equal to", equiv="<=", series_example=_le_example)
def le(df, other, axis="columns", level=None):
    op = DataFrameLessEqual(axis=axis, level=level, lhs=df, rhs=other)
    return op(df, other)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# pylint: disable=superfluous-parens,redefined-variable-type
# superfluous-parens: Sometimes extra parens are more clear

"""Bloom Filter: Probabilistic set membership testing for large sets"""

# Shamelessly borrowed (under MIT license) from
# https://code.activestate.com/recipes/577686-bloom-filter/
# About Bloom Filters: https://en.wikipedia.org/wiki/Bloom_filter

# Tweaked by Daniel Richard Stromberg, mostly to:
# 1) Give it a little nicer __init__ parameters.
# 2) Improve the hash functions to get a much lower rate of false positives.
# 3) Give it a selection of backends.
# 4) Make it pass pylint.

# In the literature:
# k is the number of probes - we call this num_probes_k
# m is the number of bits in the filter - we call this num_bits_m
# n is the ideal number of elements to eventually be stored in the filter - we
# call this ideal_num_elements_n
# p is the desired error rate when full - we call this error_rate_p

import array
import math
import os
import random

try:
    import mmap as mmap_mod
except ImportError:
    # Jython lacks mmap()
    HAVE_MMAP = False
else:
    HAVE_MMAP = True


class Mmap_backend(object):
    """
    Backend storage for our "array of bits" using an mmap'd file.
    Please note that this has only been tested on Linux so far.
    """

    effs = 2**8 - 1

    def __init__(self, num_bits, filename):
        if not HAVE_MMAP:
            raise NotImplementedError("mmap is not available")
        self.num_bits = num_bits
        self.num_chars = (self.num_bits + 7) // 8
        flags = os.O_RDWR | os.O_CREAT
        if hasattr(os, "O_BINARY"):
            flags |= getattr(os, "O_BINARY")
        self.file_ = os.open(filename, flags)
        os.lseek(self.file_, self.num_chars + 1, os.SEEK_SET)
        os.write(self.file_, b"\x00")
        self.mmap = mmap_mod.mmap(self.file_, self.num_chars)

    def is_set(self, bitno):
        """Return true iff bit number bitno is set"""
        byteno, bit_within_wordno = divmod(bitno, 8)
        mask = 1 << bit_within_wordno
        byte = self.mmap[byteno]
        return byte & mask

    def set(self, bitno):
        """set bit number bitno to true"""

        byteno, bit_within_byteno = divmod(bitno, 8)
        mask = 1 << bit_within_byteno
        byte = self.mmap[byteno]
        byte |= mask
        self.mmap[byteno] = byte

    def clear(self, bitno):
        """clear bit number bitno - set it to false"""

        byteno, bit_within_byteno = divmod(bitno, 8)
        mask = 1 << bit_within_byteno
        byte = self.mmap[byteno]
        byte &= Mmap_backend.effs - mask
        self.mmap[byteno] = byte

    def __iand__(self, other):
        assert self.num_bits == other.num_bits

        for byteno in range(self.num_chars):
            self.mmap[byteno] = self.mmap[byteno] & other.mmap[byteno]

        return self

    def __ior__(self, other):
        assert self.num_bits == other.num_bits

        for byteno in range(self.num_chars):
            self.mmap[byteno] = self.mmap[byteno] | other.mmap[byteno]

        return self

    def close(self):
        """Close the file"""
        os.close(self.file_)


class File_seek_backend(object):
    """Backend storage for our "array of bits" using a file in which we seek"""

    effs = 2**8 - 1

    def __init__(self, num_bits, filename):
        self.num_bits = num_bits
        self.num_chars = (self.num_bits + 7) // 8
        flags = os.O_RDWR | os.O_CREAT
        if hasattr(os, "O_BINARY"):
            flags |= getattr(os, "O_BINARY")
        self.file_ = os.open(filename, flags)
        os.lseek(self.file_, self.num_chars + 1, os.SEEK_SET)
        os.write(self.file_, b"\x00")

    def is_set(self, bitno):
        """Return true iff bit number bitno is set"""
        byteno, bit_within_wordno = divmod(bitno, 8)
        mask = 1 << bit_within_wordno
        os.lseek(self.file_, byteno, os.SEEK_SET)
        byte = os.read(self.file_, 1)[0]
        return byte & mask

    def set(self, bitno):
        """set bit number bitno to true"""

        byteno, bit_within_byteno = divmod(bitno, 8)
        mask = 1 << bit_within_byteno
        os.lseek(self.file_, byteno, os.SEEK_SET)
        byte = os.read(self.file_, 1)[0]
        byte |= mask
        os.lseek(self.file_, byteno, os.SEEK_SET)
        os.write(self.file_, bytes([byte]))

    def clear(self, bitno):
        """clear bit number bitno - set it to false"""

        byteno, bit_within_byteno = divmod(bitno, 8)
        mask = 1 << bit_within_byteno
        os.lseek(self.file_, byteno, os.SEEK_SET)
        byte = os.read(self.file_, 1)[0]
        byte &= File_seek_backend.effs - mask
        os.lseek(self.file_, byteno, os.SEEK_SET)
        os.write(self.file_, bytes([byte]))

    # These are quite slow ways to do iand and ior, but they should work,
    # and a faster version is going to take more time
    def __iand__(self, other):
        assert self.num_bits == other.num_bits

        for bitno in range(self.num_bits):
            if self.is_set(bitno) and other.is_set(bitno):
                self.set(bitno)
            else:
                self.clear(bitno)

        return self

    def __ior__(self, other):
        assert self.num_bits == other.num_bits

        for bitno in range(self.num_bits):
            if self.is_set(bitno) or other.is_set(bitno):
                self.set(bitno)
            else:
                self.clear(bitno)

        return self

    def close(self):
        """Close the file"""
        os.close(self.file_)


class Array_then_file_seek_backend(object):
    # pylint: disable=R0902
    # R0902: We kinda need a bunch of instance attributes
    """
    Backend storage for our "array of bits" using a python array of integers up
    to some maximum number of bytes, then spilling over to a file.
    This is -not- a cache; we instead save the leftmost bits in RAM, and the
    rightmost bits (if necessary) in a file.  On open, we read from the file to
    RAM.  On close, we write from RAM to the file.
    """

    effs = 2**8 - 1

    def __init__(self, num_bits, filename, max_bytes_in_memory):
        self.num_bits = num_bits
        num_chars = (self.num_bits + 7) // 8
        self.filename = filename
        self.max_bytes_in_memory = max_bytes_in_memory
        self.bits_in_memory = min(num_bits, self.max_bytes_in_memory * 8)
        self.bits_in_file = max(self.num_bits - self.bits_in_memory, 0)
        self.bytes_in_memory = (self.bits_in_memory + 7) // 8
        self.bytes_in_file = (self.bits_in_file + 7) // 8

        self.array_ = array.array("B", [0]) * self.bytes_in_memory
        flags = os.O_RDWR | os.O_CREAT
        if hasattr(os, "O_BINARY"):
            flags |= getattr(os, "O_BINARY")
        self.file_ = os.open(filename, flags)
        os.lseek(self.file_, num_chars + 1, os.SEEK_SET)
        os.write(self.file_, b"\x00")

        os.lseek(self.file_, 0, os.SEEK_SET)
        offset = 0
        intended_block_len = 2**17
        while True:
            if offset + intended_block_len < self.bytes_in_memory:
                block = os.read(self.file_, intended_block_len)
            elif offset < self.bytes_in_memory:
                block = os.read(self.file_, self.bytes_in_memory - offset)
            else:
                break
            for index_in_block, byte in enumerate(block):
                self.array_[offset + index_in_block] = byte
            offset += intended_block_len

    def is_set(self, bitno):
        """Return true iff bit number bitno is set"""
        byteno, bit_within_byteno = divmod(bitno, 8)
        mask = 1 << bit_within_byteno
        if byteno < self.bytes_in_memory:
            return self.array_[byteno] & mask
        else:
            os.lseek(self.file_, byteno, os.SEEK_SET)
            byte = os.read(self.file_, 1)[0]
            return byte & mask

    def set(self, bitno):
        """set bit number bitno to true"""
        byteno, bit_within_byteno = divmod(bitno, 8)
        mask = 1 << bit_within_byteno
        if byteno < self.bytes_in_memory:
            self.array_[byteno] |= mask
        else:
            os.lseek(self.file_, byteno, os.SEEK_SET)
            byte = os.read(self.file_, 1)[0]
            byte |= mask
            os.lseek(self.file_, byteno, os.SEEK_SET)
            os.write(self.file_, bytes([byte]))

    def clear(self, bitno):
        """clear bit number bitno - set it to false"""
        byteno, bit_within_byteno = divmod(bitno, 8)
        mask = Array_backend.effs - (1 << bit_within_byteno)
        if byteno < self.bytes_in_memory:
            self.array_[byteno] &= mask
        else:
            os.lseek(self.file_, byteno, os.SEEK_SET)
            byte = os.read(self.file_, 1)[0]
            byte &= File_seek_backend.effs - mask
            os.lseek(self.file_, byteno, os.SEEK_SET)
            os.write(self.file_, bytes([byte]))

    # These are quite slow ways to do iand and ior, but they should work,
    # and a faster version is going to take more time
    def __iand__(self, other):
        assert self.num_bits == other.num_bits

        for bitno in range(self.num_bits):
            if self.is_set(bitno) and other.is_set(bitno):
                self.set(bitno)
            else:
                self.clear(bitno)

        return self

    def __ior__(self, other):
        assert self.num_bits == other.num_bits

        for bitno in range(self.num_bits):
            if self.is_set(bitno) or other.is_set(bitno):
                self.set(bitno)
            else:
                self.clear(bitno)

        return self

    def close(self):
        """
        Write the in-memory portion to disk, leave the already-on-disk  portion
        unchanged
        """

        os.lseek(self.file_, 0, os.SEEK_SET)
        os.write(self.file_, bytes(self.array_[0 : self.bytes_in_memory]))

        os.close(self.file_)


class Array_backend(object):
    """
    Backend storage for our "array of bits" using a python array of integers
    """

    # Note that this has now been split out into a bits_mod for the benefit of
    # other projects.
    effs = 2**32 - 1

    def __init__(self, num_bits):
        self.num_bits = num_bits
        self.num_words = (self.num_bits + 31) // 32
        self.array_ = array.array("L", [0]) * self.num_words

    def is_set(self, bitno):
        """Return true iff bit number bitno is set"""
        wordno, bit_within_wordno = divmod(bitno, 32)
        mask = 1 << bit_within_wordno
        return self.array_[wordno] & mask

    def set(self, bitno):
        """set bit number bitno to true"""
        wordno, bit_within_wordno = divmod(bitno, 32)
        mask = 1 << bit_within_wordno
        self.array_[wordno] |= mask

    def clear(self, bitno):
        """clear bit number bitno - set it to false"""
        wordno, bit_within_wordno = divmod(bitno, 32)
        mask = Array_backend.effs - (1 << bit_within_wordno)
        self.array_[wordno] &= mask

    # It'd be nice to do __iand__ and __ior__ in a base class, but
    # that'd be Much slower

    def __iand__(self, other):
        assert self.num_bits == other.num_bits

        for wordno in range(self.num_words):
            self.array_[wordno] &= other.array_[wordno]

        return self

    def __ior__(self, other):
        assert self.num_bits == other.num_bits

        for wordno in range(self.num_words):
            self.array_[wordno] |= other.array_[wordno]

        return self

    def close(self):
        """Noop for compatibility with the file+seek backend"""
        pass


def get_bitno_seed_rnd(bloom_filter, key):
    """
    Apply num_probes_k hash functions to key.
    Generate the array index and bitmask corresponding to each result.
    """

    # We're using key as a seed to a pseudorandom number generator
    hasher = random.Random(key).randrange
    for dummy in range(bloom_filter.num_probes_k):
        bitno = hasher(bloom_filter.num_bits_m)
        yield bitno % bloom_filter.num_bits_m


MERSENNES1 = [2**x - 1 for x in [17, 31, 127]]
MERSENNES2 = [2**x - 1 for x in [19, 67, 257]]


def simple_hash(int_list, prime1, prime2, prime3):
    """Compute a hash value from a list of integers and 3 primes"""
    result = 0
    for integer in int_list:
        result += ((result + integer + prime1) * prime2) % prime3
    return result


def hash1(int_list):
    """Basic hash function #1"""
    return simple_hash(int_list, MERSENNES1[0], MERSENNES1[1], MERSENNES1[2])


def hash2(int_list):
    """Basic hash function #2"""
    return simple_hash(int_list, MERSENNES2[0], MERSENNES2[1], MERSENNES2[2])


def get_filter_bitno_probes(bloom_filter, key):
    """
    Apply num_probes_k hash functions to key.
    Generate the array index and bitmask corresponding to each result
    """

    # This one assumes key is either bytes or str (or other list of integers)

    if hasattr(key, "__divmod__"):
        int_list = []
        temp = key
        while temp:
            quotient, remainder = divmod(temp, 256)
            int_list.append(remainder)
            temp = quotient
    elif isinstance(key, (list, tuple, str, bytes)) and not key:
        int_list = []

    elif isinstance(key, (list, tuple)):
        int_list = []
        for v in key:
            if isinstance(v, str):
                int_list.extend([ord(char) for char in v])
            elif hasattr(v, "__divmod__"):
                int_list.append(v)
            else:
                raise TypeError("Sorry, I do not know how to hash this type")
    elif isinstance(key[0], str):
        int_list = [ord(char) for char in key]
    else:
        raise TypeError("Sorry, I do not know how to hash this type")

    hash_value1 = hash1(int_list)
    hash_value2 = hash2(int_list)
    probe_value = hash_value1

    for _ in range(1, bloom_filter.num_probes_k + 1):
        probe_value *= hash_value1
        probe_value += hash_value2
        probe_value %= MERSENNES1[2]
        yield probe_value % bloom_filter.num_bits_m


def try_unlink(filename):
    """unlink a file.  Don't complain if it's not there"""
    try:
        os.unlink(filename)
    except OSError:
        pass
    return


class BloomFilter(object):
    """Probabilistic set membership testing for large sets"""

    def __init__(
        self,
        max_elements=10000,
        error_rate=0.1,
        probe_bitnoer=get_filter_bitno_probes,
        filename=None,
        start_fresh=False,
    ):
        # pylint: disable=R0913
        # R0913: We want a few arguments
        if max_elements <= 0:
            raise ValueError("ideal_num_elements_n must be > 0")
        if not (0 < error_rate < 1):
            raise ValueError("error_rate_p must be between 0 and 1 exclusive")

        self.error_rate_p = error_rate
        # With fewer elements, we should do very well. With more elements, our
        # error rate "guarantee" drops rapidly.
        self.ideal_num_elements_n = max_elements

        numerator = -1 * self.ideal_num_elements_n * math.log(self.error_rate_p)
        denominator = math.log(2) ** 2
        real_num_bits_m = numerator / denominator
        self.num_bits_m = int(math.ceil(real_num_bits_m))

        if filename is None:
            self.backend = Array_backend(self.num_bits_m)
        elif isinstance(filename, tuple) and isinstance(filename[1], int):
            if start_fresh:
                try_unlink(filename[0])
            if filename[1] == -1:
                self.backend = Mmap_backend(self.num_bits_m, filename[0])
            else:
                self.backend = Array_then_file_seek_backend(
                    self.num_bits_m,
                    filename[0],
                    filename[1],
                )
        else:
            if start_fresh:
                try_unlink(filename)
            self.backend = File_seek_backend(self.num_bits_m, filename)

        # AKA num_offsetters
        # Verified against
        # https://en.wikipedia.org/wiki/Bloom_filter#Probability_of_false_positives
        real_num_probes_k = (self.num_bits_m / self.ideal_num_elements_n) * math.log(2)
        self.num_probes_k = int(math.ceil(real_num_probes_k))
        self.probe_bitnoer = probe_bitnoer

    def __repr__(self):
        return (
            "BloomFilter(ideal_num_elements_n=%d, error_rate_p=%f, " + "num_bits_m=%d)"
        ) % (
            self.ideal_num_elements_n,
            self.error_rate_p,
            self.num_bits_m,
        )

    def add(self, key):
        """Add an element to the filter"""
        for bitno in self.probe_bitnoer(self, key):
            self.backend.set(bitno)

    def __iadd__(self, key):
        self.add(key)
        return self

    def _match_template(self, bloom_filter):
        """
        Compare a sort of signature for two bloom filters.
        Used in preparation for binary operations
        """
        return (
            self.num_bits_m == bloom_filter.num_bits_m
            and self.num_probes_k == bloom_filter.num_probes_k
            and self.probe_bitnoer == bloom_filter.probe_bitnoer
        )

    def union(self, bloom_filter):
        """Compute the set union of two bloom filters"""
        self.backend |= bloom_filter.backend

    def __ior__(self, bloom_filter):
        self.union(bloom_filter)
        return self

    def intersection(self, bloom_filter):
        """Compute the set intersection of two bloom filters"""
        self.backend &= bloom_filter.backend

    def __iand__(self, bloom_filter):
        self.intersection(bloom_filter)
        return self

    def __contains__(self, key):
        for bitno in self.probe_bitnoer(self, key):
            if not self.backend.is_set(bitno):
                return False
        return True

    def close(self):
        self.backend.close()
        self.backend = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        self.backend = None

    def __del__(self):
        if self.backend is not None:
            self.backend.close()
            self.backend = None



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import importlib
import inspect
import os
from abc import ABC, ABCMeta, abstractmethod
from typing import Any, Callable, Optional, Type, Union

from ...utils import no_default
from .field_type import (
    AbstractFieldType,
    FieldTypes,
    ListType,
    TupleType,
    DictType,
    ReferenceType,
)

_is_ci = (os.environ.get("CI") or "0").lower() in ("1", "true")


class Field(ABC):
    __slots__ = (
        "_tag",
        "_default_value",
        "_default_factory",
        "_on_serialize",
        "_on_deserialize",
        "name",  # The __name__ of member_descriptor
        "get",  # The __get__ of member_descriptor
        "set",  # The __set__ of member_descriptor
        "__delete__",  # The __delete__ of member_descriptor
    )

    _tag: str
    _default_value: Any
    _default_factory: Optional[Callable]

    def __init__(
        self,
        tag: str,
        default: Any = no_default,
        default_factory: Optional[Callable] = None,
        on_serialize: Callable[[Any], Any] = None,
        on_deserialize: Callable[[Any], Any] = None,
    ):
        if (
            default is not no_default and default_factory is not None
        ):  # pragma: no cover
            raise ValueError("default and default_factory can not be specified both")

        self._tag = tag
        self._default_value = default
        self._default_factory = default_factory
        self._on_serialize = on_serialize
        self._on_deserialize = on_deserialize

    @property
    def tag(self):
        return self._tag

    @property
    def on_serialize(self):
        return self._on_serialize

    @property
    def on_deserialize(self):
        return self._on_deserialize

    @property
    @abstractmethod
    def field_type(self) -> AbstractFieldType:
        """
        Field type.

        Returns
        -------
        field_type : AbstractFieldType
             Field type.
        """

    def __get__(self, instance, owner=None):
        try:
            return self.get(instance, owner)
        except AttributeError:
            if self._default_value is not no_default:
                val = self._default_value
                self.set(instance, val)
                return val
            elif self._default_factory is not None:
                val = self._default_factory()
                self.set(instance, val)
                return val
            else:
                raise

    def __set__(self, instance, value) -> None:
        if _is_ci:  # pragma: no branch
            from ...core import is_kernel_mode

            if not is_kernel_mode():
                field_type = self.field_type
                try:
                    to_check_value = value
                    if to_check_value is not None and self._on_serialize:
                        to_check_value = self._on_serialize(to_check_value)
                    field_type.validate(to_check_value)
                except (TypeError, ValueError) as e:
                    raise type(e)(
                        f"Failed to set `{self.name}` for {type(instance).__name__} "
                        f"when environ CI=true is set: {str(e)}"
                    )
        self.set(instance, value)


class AnyField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.any


class IdentityField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.string


class BoolField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.bool


class Int8Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.int8


class Int16Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.int16


class Int32Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.int32


class Int64Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.int64


class UInt8Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.uint8


class UInt16Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.uint16


class UInt32Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.uint32


class UInt64Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.uint64


class Float16Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.float16


class Float32Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.float32


class Float64Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.float64


class Complex64Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.complex64


class Complex128Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.complex128


class StringField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.string


class BytesField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.bytes


class KeyField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.key


class NDArrayField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.ndarray


class Datetime64Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.datetime


class Timedelta64Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.timedelta


class DataTypeField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.dtype


class IndexField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.index


class SeriesField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.series


class DataFrameField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.dataframe


class SliceField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.slice


class FunctionField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.function


class NamedTupleField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.namedtuple


class TZInfoField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.tzinfo


class IntervalArrayField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.interval_array


class _CollectionField(Field, metaclass=ABCMeta):
    __slots__ = ("_field_type",)

    def __init__(
        self,
        tag: str,
        field_type: AbstractFieldType = None,
        default: Any = no_default,
        default_factory: Optional[Callable] = None,
        on_serialize: Callable[[Any], Any] = None,
        on_deserialize: Callable[[Any], Any] = None,
    ):
        super().__init__(
            tag,
            default=default,
            default_factory=default_factory,
            on_serialize=on_serialize,
            on_deserialize=on_deserialize,
        )
        if field_type is None:
            field_type = FieldTypes.any
        if not isinstance(field_type, ListType):
            field_type = self._collection_type()(field_type, ...)
        self._field_type = field_type

    @classmethod
    @abstractmethod
    def _collection_type(cls) -> AbstractFieldType:
        """
        Collection type.

        Returns
        -------
        collection_type
        """

    @property
    def field_type(self) -> Type[AbstractFieldType]:
        return self._field_type


class ListField(_CollectionField):
    __slots__ = ()

    @classmethod
    def _collection_type(cls) -> Type[AbstractFieldType]:
        return ListType


class TupleField(_CollectionField):
    __slots__ = ()

    @classmethod
    def _collection_type(cls) -> Type[AbstractFieldType]:
        return TupleType


class DictField(Field):
    __slots__ = ("_field_type",)

    def __init__(
        self,
        tag: str,
        key_type: AbstractFieldType = None,
        value_type: AbstractFieldType = None,
        default: Any = no_default,
        default_factory: Optional[Callable] = None,
        on_serialize: Callable[[Any], Any] = None,
        on_deserialize: Callable[[Any], Any] = None,
    ):
        super().__init__(
            tag,
            default=default,
            default_factory=default_factory,
            on_serialize=on_serialize,
            on_deserialize=on_deserialize,
        )
        self._field_type = DictType(key_type, value_type)

    @property
    def field_type(self) -> AbstractFieldType:
        return self._field_type


class ReferenceField(Field):
    __slots__ = "_reference_type", "_field_type"

    def __init__(
        self,
        tag: str,
        reference_type: Union[str, Type] = None,
        default: Any = no_default,
        on_serialize: Callable[[Any], Any] = None,
        on_deserialize: Callable[[Any], Any] = None,
    ):
        super().__init__(
            tag,
            default=default,
            on_serialize=on_serialize,
            on_deserialize=on_deserialize,
        )
        self._reference_type = reference_type

        if not isinstance(reference_type, str):
            self._field_type = ReferenceType(reference_type)
        else:
            # need to bind dynamically
            self._field_type = None

    @property
    def field_type(self) -> AbstractFieldType:
        return self._field_type

    def get_field_type(self, instance):
        if self._field_type is None:
            # bind dynamically
            if self._reference_type == "self":
                reference_type = type(instance)
            elif isinstance(self._reference_type, str) and "." in self._reference_type:
                module, name = self._reference_type.rsplit(".", 1)
                reference_type = getattr(importlib.import_module(module), name)
            else:
                module = inspect.getmodule(instance)
                reference_type = getattr(module, self._reference_type)
            self._field_type = ReferenceType(reference_type)
        return self._field_type

    def __set__(self, instance, value):
        if _is_ci:
            from ...core import is_kernel_mode

            if not is_kernel_mode():
                field_type = self.get_field_type(instance)
                try:
                    to_check_value = value
                    if to_check_value is not None and self._on_serialize:
                        to_check_value = self._on_serialize(to_check_value)
                    field_type.validate(to_check_value)
                except (TypeError, ValueError) as e:
                    raise type(e)(
                        f"Failed to set `{self.name}` for {type(instance).__name__} "
                        f"when environ CI=true is set: {e}"
                    )
        self.set(instance, value)


class OneOfField(Field):
    __slots__ = "_reference_fields"

    def __init__(
        self,
        tag: str,
        default: Any = no_default,
        on_serialize: Callable[[Any], Any] = None,
        on_deserialize: Callable[[Any], Any] = None,
        **tag_to_reference_types,
    ):
        super().__init__(
            tag,
            default=default,
            on_serialize=on_serialize,
            on_deserialize=on_deserialize,
        )
        self._reference_fields = [
            ReferenceField(t, ref_type)
            for t, ref_type in tag_to_reference_types.items()
        ]

    @property
    def reference_fields(self):
        return self._reference_fields

    @property
    def field_type(self) -> AbstractFieldType:  # pragma: no cover
        # takes no effect here, just return AnyType()
        # we will do check in __set__ instead
        return FieldTypes.any

    def __set__(self, instance, value):
        if not _is_ci:  # pragma: no cover
            return self.set(instance, value)

        for reference_field in self._reference_fields:
            try:
                to_check_value = value
                if to_check_value is not None and self._on_serialize:
                    to_check_value = self._on_serialize(to_check_value)
                reference_field.get_field_type(instance).validate(to_check_value)
                self.set(instance, value)
                return
            except TypeError:
                continue
        valid_types = list(
            itertools.chain(
                *[
                    r.get_field_type(instance).valid_types
                    for r in self._reference_fields
                ]
            )
        )
        raise TypeError(
            f"Failed to set `{self.name}` for {type(instance).__name__} "
            f"when environ CI=true is set: type of instance cannot match any "
            f"of {valid_types}, got {type(value).__name__}"
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ...serialization.serializables import ReferenceField
from .chunks import ChunkData, Chunk, CHUNK_TYPE


class FuseChunkData(ChunkData):
    __slots__ = ("_inited",)

    _chunk = ReferenceField(
        "chunk", CHUNK_TYPE, on_serialize=lambda x: x.data if hasattr(x, "data") else x
    )

    def __init__(self, *args, **kwargs):
        self._inited = False
        super().__init__(*args, **kwargs)
        self._extra_params = {}
        self._inited = True

    @property
    def chunk(self):
        return self._chunk

    @property
    def composed(self):
        # for compatibility, just return the topological ordering,
        # once we apply optimization on the subgraph,
        # `composed` is not needed any more and should be removed then.
        assert getattr(self._op, "fuse_graph", None) is not None
        fuse_graph = self._op.fuse_graph
        return list(fuse_graph.topological_iter())

    def __getattr__(self, attr):
        if not self._inited:
            return object.__getattribute__(self, attr)
        if attr in self._extra_params:
            return self._extra_params[attr]
        try:
            return getattr(self._chunk, attr)
        except AttributeError:
            return object.__getattribute__(self, attr)

    def __setattr__(self, attr, value):
        if attr == "params":
            self._chunk.params = value
        else:
            super().__setattr__(attr, value)

    @property
    def nbytes(self):
        return np.prod(self.shape) * self.dtype.itemsize


class FuseChunk(Chunk):
    __slots__ = ()
    _allow_data_type_ = (FuseChunkData,)


FUSE_CHUNK_TYPE = (FuseChunkData, FuseChunk)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterator, Optional, Sized, Sequence
import math

try:
    import torch
    from torch.utils.data import Sampler
except ImportError:  # pragma: no cover
    torch = None
    Sampler = object

from ....utils import require_not_none


@require_not_none(torch)
class SequentialSampler(Sampler):
    r"""
    Samples elements sequentially, always in the same order.

    Args:
        data_source (Dataset): dataset to sample from
    """
    data_source: Sized

    def __init__(self, data_source):
        self.data_source = data_source

    def __iter__(self) -> Iterator[int]:
        return iter(range(len(self.data_source)))

    def __len__(self) -> int:
        return len(self.data_source)


@require_not_none(torch)
class RandomSampler(Sampler):
    r"""
    Samples elements randomly. If without replacement, then sample from a shuffled dataset.
    If with replacement, then user can specify :attr:`num_samples` to draw.

    Args:
        data_source (Dataset): dataset to sample from
        replacement (bool): samples are drawn on-demand with replacement if ``True``, default=``False``
        num_samples (int): number of samples to draw, default=`len(dataset)`. This argument
            is supposed to be specified only when `replacement` is ``True``.
        generator (Generator): Generator used in sampling.
    """
    data_source: Sized
    replacement: bool

    def __init__(
        self, data_source, replacement=False, num_samples=None, generator=None
    ):

        self.data_source = data_source
        self.replacement = replacement
        self._num_samples = num_samples
        self.generator = generator

        if not isinstance(self.replacement, bool):
            raise ValueError(
                "replacement should be a boolean value, but got "
                f"replacement={self.replacement}"
            )

        if self._num_samples is not None and not replacement:
            raise ValueError(
                "With replacement=False, num_samples should not be specified, "
                "since a random permute will be performed."
            )

        if not isinstance(self.num_samples, int) or self.num_samples <= 0:
            raise ValueError(
                "num_samples should be a positive integer "
                f"value, but got num_samples={self.num_samples}"
            )

    @property
    def num_samples(self):
        # dataset size might change at runtime
        if self._num_samples is None:
            return len(self.data_source)
        return self._num_samples

    def __iter__(self):
        n = len(self.data_source)
        if self.generator is None:
            generator = torch.Generator()
            generator.manual_seed(
                int(torch.empty((), dtype=torch.int64).random_().item())
            )
        else:
            generator = self.generator
        if self.replacement:
            for _ in range(self.num_samples // 32):
                yield from torch.randint(
                    high=n, size=(32,), dtype=torch.int64, generator=generator
                ).tolist()
            yield from torch.randint(
                high=n,
                size=(self.num_samples % 32,),
                dtype=torch.int64,
                generator=generator,
            ).tolist()
        else:
            yield from torch.randperm(n, generator=generator).tolist()

    def __len__(self) -> int:
        return self.num_samples


@require_not_none(torch)
class SubsetRandomSampler(Sampler):
    """
    Samples elements randomly from a given list of indices, without replacement.

    Args:
        indices (sequence): a sequence of indices
        generator (Generator): Generator used in sampling.
    """

    indices: Sequence[int]

    def __init__(self, indices: Sequence[int], generator=None) -> None:
        self.indices = indices
        self.generator = generator

    def __iter__(self) -> Iterator[int]:
        return (
            self.indices[i]
            for i in torch.randperm(len(self.indices), generator=self.generator)
        )

    def __len__(self) -> int:
        return len(self.indices)


@require_not_none(torch)
class DistributedSampler(Sampler):
    r"""
    Sampler that restricts data loading to a subset of the dataset.

    It is especially useful in conjunction with
    :class:`torch.nn.parallel.DistributedDataParallel`. In such a case, each
    process can pass a :class:`~torch.utils.data.DistributedSampler` instance as a
    :class:`~torch.utils.data.DataLoader` sampler, and load a subset of the
    original dataset that is exclusive to it.

    .. note::
        Dataset is assumed to be of constant size.

    Args:
        dataset: Dataset used for sampling.
        num_replicas (int, optional): Number of processes participating in
            distributed training. By default, :attr:`world_size` is retrieved from the
            current distributed group.
        rank (int, optional): Rank of the current process within :attr:`num_replicas`.
            By default, :attr:`rank` is retrieved from the current distributed
            group.
        shuffle (bool, optional): If ``True`` (default), sampler will shuffle the
            indices.
        seed (int, optional): random seed used to shuffle the sampler if
            :attr:`shuffle=True`. This number should be identical across all
            processes in the distributed group. Default: ``0``.
        drop_last (bool, optional): if ``True``, then the sampler will drop the
            tail of the data to make it evenly divisible across the number of
            replicas. If ``False``, the sampler will add extra indices to make
            the data evenly divisible across the replicas. Default: ``False``.

    .. warning::
        In distributed mode, calling the :meth:`set_epoch` method at
        the beginning of each epoch **before** creating the :class:`DataLoader` iterator
        is necessary to make shuffling work properly across multiple epochs. Otherwise,
        the same ordering will be always used.

    Example::

        >>> sampler = DistributedSampler(dataset) if is_distributed else None
        >>> loader = DataLoader(dataset, shuffle=(sampler is None),
        ...                     sampler=sampler)
        >>> for epoch in range(start_epoch, n_epochs):
        ...     if is_distributed:
        ...         sampler.set_epoch(epoch)
        ...     train(loader)
    """

    def __init__(
        self,
        dataset,
        num_replicas: Optional[int] = None,
        rank: Optional[int] = None,
        shuffle: bool = True,
        seed: int = 0,
        drop_last: bool = False,
    ) -> None:
        import torch.distributed as dist

        if num_replicas is None:  # pragma: no cover
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = dist.get_world_size()
        if rank is None:  # pragma: no cover
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = dist.get_rank()
        if rank >= num_replicas or rank < 0:
            raise ValueError(
                "Invalid rank {}, rank should be in the interval"
                " [0, {}]".format(rank, num_replicas - 1)
            )

        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0
        self.drop_last = drop_last
        # If the dataset length is evenly divisible by # of replicas, then there
        # is no need to drop any data, since the dataset will be split equally.
        if self.drop_last and len(self.dataset) % self.num_replicas != 0:
            # Split to nearest available length that is evenly divisible.
            # This is to ensure each rank receives the same amount of data when
            # using this Sampler.
            self.num_samples = math.ceil(
                (len(self.dataset) - self.num_replicas) / self.num_replicas
            )
        else:
            self.num_samples = math.ceil(len(self.dataset) / self.num_replicas)
        self.total_size = self.num_samples * self.num_replicas
        self.shuffle = shuffle
        self.seed = seed

    def generate_indices(self):
        if self.shuffle:
            # deterministically shuffle based on epoch and seed
            g = torch.Generator()
            g.manual_seed(self.seed + self.epoch)
            indices = torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = list(range(len(self.dataset)))

        if not self.drop_last:
            # add extra samples to make it evenly divisible
            padding_size = self.total_size - len(indices)
            if padding_size <= len(indices):
                indices += indices[:padding_size]
            else:
                indices += (indices * math.ceil(padding_size / len(indices)))[
                    :padding_size
                ]
        else:
            # remove tail of data to make it evenly divisible.
            indices = indices[: self.total_size]
        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank : self.total_size : self.num_replicas]
        assert len(indices) == self.num_samples

        return indices

    def __iter__(self):
        return iter(self.generate_indices())

    def __len__(self):
        return self.num_samples

    def set_epoch(self, epoch: int):
        r"""Sets the epoch for this sampler. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.

        Args:
            epoch (int): Epoch number.
        """
        self.epoch = epoch



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..datasource import tensor as astensor
from .cholesky import cholesky
from .lu import lu
from .solve_triangular import solve_triangular


def solve(a, b, sym_pos=False, sparse=None):
    """
    Solve the equation ``a x = b`` for ``x``.

    Parameters
    ----------
    a : (M, M) array_like
        A square matrix.
    b : (M,) or (M, N) array_like
        Right-hand side matrix in ``a x = b``.
    sym_pos : bool
        Assume `a` is symmetric and positive definite. If ``True``, use Cholesky
        decomposition.
    sparse: bool, optional
        Return sparse value or not.

    Returns
    -------
    x : (M,) or (M, N) ndarray
    Solution to the system ``a x = b``.  Shape of the return matches the
    shape of `b`.

    Raises
    ------
    LinAlgError
    If `a` is singular.

    Examples
    --------
    Given `a` and `b`, solve for `x`:

    >>> import mars.tensor as mt
    >>> a = mt.array([[3, 2, 0], [1, -1, 0], [0, 5, 1]])
    >>> b = mt.array([2, 4, -1])
    >>> x = mt.linalg.solve(a, b)
    >>> x.execute()
    array([ 2., -2.,  9.])

    >>> mt.dot(a, x).execute()  # Check the result
    array([ 2., 4., -1.])
    """
    a = astensor(a)
    b = astensor(b)
    if sym_pos:
        l_ = cholesky(a, lower=True)
        u = l_.T
    else:
        p, l_, u = lu(a)
        b = p.T.dot(b)
    sparse = sparse if sparse is not None else a.issparse()
    uy = solve_triangular(l_, b, lower=True, sparse=sparse)
    return solve_triangular(u, uy, sparse=sparse)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
from typing import Type

from .... import oscar as mo
from ....oscar.backends.allocate_strategy import IdleLabel
from .runner import SubtaskRunnerActor


class SubtaskRunnerManagerActor(mo.Actor):
    def __init__(self, worker_address: str, subtask_processor_cls: Type):
        # specify subtask process class
        # for test purpose
        self._worker_address = worker_address
        self._subtask_processor_cls = subtask_processor_cls
        self._cluster_api = None

        self._band_slot_runner_refs = dict()

    async def __post_create__(self):
        from ...cluster.api import ClusterAPI

        self._cluster_api = await ClusterAPI.create(self.address)

        band_to_resource = await self._cluster_api.get_bands()
        for band, resource in band_to_resource.items():
            await self._create_band_runner_actors(
                band[1], int(resource.num_cpus or resource.num_gpus)
            )

    async def _create_band_runner_actors(self, band_name: str, n_slots: int):
        strategy = IdleLabel(band_name, "subtask_runner")
        band = (self.address, band_name)
        for slot_id in range(n_slots):
            self._band_slot_runner_refs[(band_name, slot_id)] = await mo.create_actor(
                SubtaskRunnerActor,
                band,
                worker_address=self._worker_address,
                subtask_processor_cls=self._subtask_processor_cls,
                uid=SubtaskRunnerActor.gen_uid(band_name, slot_id),
                address=self.address,
                allocate_strategy=strategy,
            )

    async def __pre_destroy__(self):
        await asyncio.gather(
            *[mo.destroy_actor(ref) for ref in self._band_slot_runner_refs.values()]
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import os
import time
from typing import Dict, List, NamedTuple, Set, Tuple

import psutil

from .... import oscar as mo
from ....oscar.errors import NoFreeSlot, SlotStateError
from ....oscar.backends.allocate_strategy import IdleLabel
from ....resource import Resource
from ....typing import BandType
from ...cluster import WorkerSlotInfo, ClusterAPI

logger = logging.getLogger(__name__)


class DispatchDumpType(NamedTuple):
    free_slots: Set
    fresh_slots: Set


class WorkerSlotManagerActor(mo.Actor):
    _band_slot_infos: Dict[str, List[WorkerSlotInfo]]

    def __init__(self):
        self._cluster_api = None
        self._global_resource_ref = None

        self._band_slot_managers = dict()  # type: Dict[str, mo.ActorRef]

    async def __post_create__(self):
        self._cluster_api = await ClusterAPI.create(self.address)

        band_to_resource = await self._cluster_api.get_bands()
        for band, resource in band_to_resource.items():
            self._band_slot_managers[band] = await mo.create_actor(
                BandSlotManagerActor,
                band,
                int(resource.num_cpus or resource.num_gpus),
                self._global_resource_ref,
                uid=BandSlotManagerActor.gen_uid(band[1]),
                address=self.address,
            )

    async def __pre_destroy__(self):
        await asyncio.gather(
            *[mo.destroy_actor(ref) for ref in self._band_slot_managers.values()]
        )


class BandSlotManagerActor(mo.Actor):
    _free_slots: Set[int]
    _fresh_slots: Set[int]

    @classmethod
    def gen_uid(cls, band_name: str):
        return f"{band_name}_band_slot_manager"

    def __init__(
        self, band: BandType, n_slots: int, global_resource_ref: mo.ActorRef = None
    ):
        super().__init__()
        self._cluster_api = None

        self._band = band
        self._band_name = band[1]
        self._global_resource_ref = global_resource_ref
        self._n_slots = n_slots

        self._semaphore = asyncio.Semaphore(0)
        self._slot_control_refs = dict()
        self._free_slots = set()
        self._fresh_slots = set()
        self._slot_kill_events = dict()

        self._restarting = False
        self._restart_done_event = asyncio.Event()

        self._session_stid_to_slot = dict()
        self._slot_to_session_stid = dict()
        self._last_report_time = time.time()

        self._slot_to_proc = dict()
        self._usage_upload_task = None

    async def __post_create__(self):
        try:
            self._cluster_api = await ClusterAPI.create(self.address)
        except mo.ActorNotExist:
            pass

        strategy = IdleLabel(self._band_name, f"worker_slot_control")
        for slot_id in range(self._n_slots):
            self._slot_control_refs[slot_id] = await mo.create_actor(
                BandSlotControlActor,
                self.ref(),
                self._band_name,
                slot_id,
                uid=BandSlotControlActor.gen_uid(self._band_name, slot_id),
                address=self.address,
                allocate_strategy=strategy,
            )
            self._fresh_slots.add(slot_id)

        self._upload_slot_usage_with_delay()

    async def __pre_destroy__(self):
        self._usage_upload_task.cancel()

    async def _get_global_resource_ref(self):
        if self._global_resource_ref is not None:
            return self._global_resource_ref

        from ..supervisor import GlobalResourceManagerActor

        try:
            [self._global_resource_ref] = await self._cluster_api.get_supervisor_refs(
                [GlobalResourceManagerActor.default_uid()]
            )
        except mo.ActorNotExist:
            self._global_resource_ref = None
        return self._global_resource_ref

    def get_slot_address(self, slot_id: int):
        return self._slot_control_refs[slot_id].address

    def get_subtask_slot(self, session_stid: Tuple[str, str]):
        return self._session_stid_to_slot.get(session_stid)

    async def acquire_free_slot(self, session_stid: Tuple[str, str], block=True):
        if not block and self._semaphore.locked():
            raise NoFreeSlot(f"No free slot for {session_stid}")
        yield self._semaphore.acquire()
        if self._restarting:
            yield self._restart_done_event.wait()

        slot_id = self._free_slots.pop()
        self._fresh_slots.difference_update([slot_id])
        self._slot_to_session_stid[slot_id] = session_stid
        self._session_stid_to_slot[session_stid] = slot_id
        logger.debug("Slot %d acquired for subtask %r", slot_id, session_stid)
        raise mo.Return(slot_id)

    def release_free_slot(self, slot_id: int, session_stid: Tuple[str, str]):
        acquired_session_stid = self._slot_to_session_stid.pop(slot_id, None)
        if acquired_session_stid is None:
            raise SlotStateError(f"Slot {slot_id} is not acquired.")
        if acquired_session_stid != session_stid:
            raise SlotStateError(
                f"Slot {slot_id} releasing state incorrect, "
                f"the acquired session_stid: {acquired_session_stid}, "
                f"the releasing session_stid: {session_stid}"
            )
        acquired_slot_id = self._session_stid_to_slot.pop(acquired_session_stid)
        assert (
            acquired_slot_id == slot_id
        ), f"{acquired_session_stid}: acquired_slot_id {acquired_slot_id} != slot_id {slot_id}"

        logger.debug("Slot %d released", slot_id)

        if slot_id not in self._free_slots:
            self._free_slots.add(slot_id)
            self._semaphore.release()

    def register_slot(self, slot_id: int, pid: int):
        try:
            self._fresh_slots.add(slot_id)
            if slot_id in self._slot_kill_events:
                event = self._slot_kill_events.pop(slot_id)
                event.set()
            if slot_id in self._slot_to_session_stid:
                # We should release the slot by one role, if the slot is
                # acquired by the SubtaskExecutionActor, then the slot
                # should be released by it, too.
                session_stid = self._slot_to_session_stid[slot_id]
                logger.info(
                    "Slot %s registered by pid %s, current acquired session_stid is %s",
                    slot_id,
                    pid,
                    session_stid,
                )
            else:
                if slot_id not in self._free_slots:
                    self._free_slots.add(slot_id)
                    self._semaphore.release()
        finally:
            # psutil may raises exceptions, but currently we can't handle the register exception,
            # so put it to the finally.
            # TODO(fyrestone): handle register_slot failure.
            self._slot_to_proc[slot_id] = proc = psutil.Process(pid)
            # collect initial stats for the process
            proc.cpu_percent(interval=None)

    async def _kill_slot(self, slot_id: int):
        if slot_id in self._slot_kill_events:
            await self._slot_kill_events[slot_id].wait()
            return

        event = self._slot_kill_events[slot_id] = asyncio.Event()
        # TODO(fyrestone): Make it more reliable. e.g. kill_actor
        # success but the actor does not restart.
        try:
            await mo.kill_actor(self._slot_control_refs[slot_id])
        except ConnectionError:
            pass
        await event.wait()

    async def kill_slot(self, slot_id: int):
        self._free_slots.difference_update([slot_id])
        yield self._kill_slot(slot_id)

    async def restart_free_slots(self):
        if self._restarting:
            yield self._restart_done_event.wait()
            return

        self._restart_done_event = asyncio.Event()
        self._restarting = True
        slot_ids = [
            slot_id for slot_id in self._free_slots if slot_id not in self._fresh_slots
        ]
        if slot_ids:
            yield asyncio.gather(*[self._kill_slot(slot_id) for slot_id in slot_ids])
            logger.info("%d idle slots restarted", len(slot_ids))

        self._restarting = False
        self._restart_done_event.set()

    def _upload_slot_usage_with_delay(self, delay: int = 1):
        self._usage_upload_task = self.ref().upload_slot_usages.tell_delay(
            periodical=True, delay=delay
        )

    async def upload_slot_usages(self, periodical: bool = False):
        delays = []
        slot_infos = []
        global_resource_ref = await self._get_global_resource_ref()

        if global_resource_ref is None:  # pragma: no cover
            if periodical:
                self._upload_slot_usage_with_delay()
            return

        for slot_id, proc in self._slot_to_proc.items():
            if slot_id not in self._slot_to_session_stid:
                continue
            session_id, subtask_id = self._slot_to_session_stid[slot_id]
            cpu_usage, gpu_usage, processor_usage = 0, 0, 0
            if self._band_name.startswith("gpu"):
                processor_usage = gpu_usage = 1
            else:
                try:
                    processor_usage = cpu_usage = (
                        proc.cpu_percent(interval=None) / 100.0
                    )
                except psutil.NoSuchProcess:  # pragma: no cover
                    continue
                except psutil.AccessDenied as e:  # pragma: no cover
                    logger.warning("Access denied when getting cpu percent: %s", e)
                    processor_usage = cpu_usage = 0.0

            slot_infos.append(
                WorkerSlotInfo(
                    slot_id=slot_id,
                    session_id=session_id,
                    subtask_id=subtask_id,
                    processor_usage=processor_usage,
                )
            )

            if global_resource_ref is not None:  # pragma: no branch
                # FIXME fix band slot mistake
                delays.append(
                    global_resource_ref.update_subtask_resources.delay(
                        self._band[1],
                        session_id,
                        subtask_id,
                        Resource(
                            num_cpus=max(1.0, cpu_usage), num_gpus=max(1.0, gpu_usage)
                        ),
                    )
                )

        if delays:  # pragma: no branch
            yield global_resource_ref.update_subtask_resources.batch(*delays)
        if self._cluster_api is not None:
            await self._cluster_api.set_band_slot_infos(self._band_name, slot_infos)

        if periodical:
            self._upload_slot_usage_with_delay()

    def dump_data(self):
        """
        Get all refs of slots of a queue
        """
        return DispatchDumpType(self._free_slots, self._fresh_slots)


class BandSlotControlActor(mo.Actor):
    @classmethod
    def gen_uid(cls, band_name: str, slot_id: int):
        return f"{band_name}_{slot_id}_band_slot_control"

    def __init__(self, manager_ref, band_name, slot_id: int):
        self._manager_ref = manager_ref
        self._band_name = band_name
        self._slot_id = slot_id
        self._report_task = None

    async def __post_create__(self):
        self._report_task = asyncio.create_task(self._report_slot_ready())

    async def _report_slot_ready(self):
        from ...cluster.api import ClusterAPI

        try:
            self._cluster_api = await ClusterAPI.create(self.address)
            await self._cluster_api.wait_node_ready()
        except mo.ActorNotExist:
            pass

        await mo.wait_actor_pool_recovered(self.address)
        await self._manager_ref.register_slot.tell(self._slot_id, os.getpid())



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import threading
import multiprocessing
from concurrent.futures import Executor
from typing import Union

from .base import AioBase, delegate_to_executor, proxy_method_directly

event_types = Union[threading.Event, multiprocessing.Event]


@delegate_to_executor("wait")
@proxy_method_directly("set", "is_set", "clear")
class AioEvent(AioBase):
    def __init__(
        self,
        event: event_types = None,
        loop: asyncio.BaseEventLoop = None,
        executor: Executor = None,
    ):
        if event is None:
            event = threading.Event()
        super().__init__(event, loop=loop, executor=executor)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...tensor.arithmetic import sqrt
from .var import var_dataframe, var_series


def std_dataframe(
    df,
    axis=None,
    skipna=True,
    level=None,
    ddof=1,
    numeric_only=None,
    combine_size=None,
    method=None,
):
    ret = sqrt(
        var_dataframe(
            df,
            axis=axis,
            skipna=skipna,
            level=level,
            ddof=ddof,
            numeric_only=numeric_only,
            combine_size=combine_size,
            method=method,
        )
    )
    return ret


def std_series(
    series, axis=None, skipna=True, level=None, ddof=1, combine_size=None, method=None
):
    ret = sqrt(
        var_series(
            series,
            axis=axis,
            skipna=skipna,
            level=level,
            ddof=ddof,
            combine_size=combine_size,
            method=method,
        )
    )
    return ret



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import os

from ..base import path_type, stringify_path
from ....utils import lazy_import

oss2 = lazy_import("oss2", placeholder=True)

# OSS api time out
_oss_time_out = 10


class OSSFileEntry:
    def __init__(
        self, path, *, is_dir=None, is_file=None, stat=None, storage_options=None
    ):
        self._path = path
        self._name = os.path.basename(path)
        self._is_file = is_file
        self._is_dir = is_dir
        self._stat = stat
        self._storage_options = storage_options

    def is_dir(self):
        if self._is_dir is None:
            self._is_dir = oss_isdir(self._path)
        return self._is_dir

    def is_file(self):
        if self._is_file is None:
            if self.is_dir() or not oss_exists(self._path):
                self._is_file = False
            else:
                self._is_file = True
        return self._is_file

    def stat(self):
        if self._stat is None:
            self._stat = oss_stat(self._path)
        return self._stat

    @property
    def name(self):
        return self._name

    @property
    def path(self):
        return self._path


def parse_osspath(path: path_type):
    # Extract OSS configuration from the encoded URL.
    str_path = stringify_path(path)
    parse_result = oss2.urlparse(str_path)
    if parse_result.scheme != "oss":
        raise ValueError(
            f"Except scheme oss, but got scheme: {parse_result.scheme}"
            f" in path: {str_path}"
        )
    bucket = parse_result.hostname
    if not (parse_result.username and parse_result.password):
        raise RuntimeError(r"Please use build_oss_path to add OSS info")
    param_dict = url_to_dict(parse_result.username)
    access_key_id = param_dict["access_key_id"]
    access_key_secret = parse_result.password
    end_point = param_dict["end_point"]
    key = parse_result.path
    key = key[1:] if key.startswith("/") else key
    return bucket, key, access_key_id, access_key_secret, end_point


def _get_oss_bucket(bucket, access_key_id, access_key_secret, end_point):
    oss_bucket = oss2.Bucket(
        auth=oss2.Auth(
            access_key_id=access_key_id, access_key_secret=access_key_secret
        ),
        endpoint=end_point,
        bucket_name=bucket,
        connect_timeout=_oss_time_out,
    )
    return oss_bucket


def oss_exists(path: path_type):
    bucket, key, access_key_id, access_key_secret, end_point = parse_osspath(path)
    oss_bucket = _get_oss_bucket(bucket, access_key_id, access_key_secret, end_point)
    return oss_bucket.object_exists(key) or oss_isdir(path)


def oss_isdir(path: path_type):
    """
    OSS has no concept of directories, but we define
    a ossurl is dir, When there is at least one object
    at the ossurl that is the prefix(end with char "/"),
    it is considered as a directory.
    """
    dirname = stringify_path(path)
    if not dirname.endswith("/"):
        dirname = dirname + "/"
    bucket, key, access_key_id, access_key_secret, end_point = parse_osspath(dirname)
    oss_bucket = _get_oss_bucket(bucket, access_key_id, access_key_secret, end_point)
    isdir = False
    for obj in oss2.ObjectIteratorV2(oss_bucket, prefix=key, max_keys=2):
        if obj.key == key:
            continue
        isdir = True
        break
    return isdir


def oss_stat(path: path_type):
    path = stringify_path(path)
    bucket, key, access_key_id, access_key_secret, end_point = parse_osspath(path)
    oss_bucket = _get_oss_bucket(bucket, access_key_id, access_key_secret, end_point)
    if oss_isdir(path):
        stat = dict(name=path, size=0, modified_time=-1)
        stat["type"] = "directory"
    else:
        meta = oss_bucket.get_object_meta(key)
        stat = dict(
            name=path,
            size=int(meta.headers["Content-Length"]),
            modified_time=meta.headers["Last-Modified"],
        )
        stat["type"] = "file"
    return stat


def oss_scandir(dirname: path_type):
    dirname = stringify_path(dirname)
    if not dirname.endswith("/"):
        dirname = dirname + "/"
    bucket, key, access_key_id, access_key_secret, end_point = parse_osspath(dirname)
    oss_bucket = _get_oss_bucket(bucket, access_key_id, access_key_secret, end_point)
    dirname_set = set()
    for obj in oss2.ObjectIteratorV2(oss_bucket, prefix=key):
        rel_path = obj.key[len(key) :]
        try:
            inside_dirname, inside_filename = rel_path.split("/", 1)
        except ValueError:
            inside_dirname = None
            inside_filename = rel_path
        if inside_dirname is not None:
            if inside_dirname in dirname_set:
                continue
            dirname_set.add(inside_dirname)
            yield OSSFileEntry(
                os.path.join(dirname, inside_dirname),
                is_dir=True,
                is_file=False,
                stat={
                    "name": os.path.join(dirname, inside_dirname),
                    "type": "directory",
                    "size": 0,
                    "modified_time": -1,
                },
            )
        else:
            yield OSSFileEntry(
                os.path.join(dirname, inside_filename),
                is_dir=False,
                is_file=True,
                stat={
                    "name": os.path.join(dirname, inside_filename),
                    "type": "file",
                    "size": obj.size,
                    "modified_time": obj.last_modified,
                },
            )


def dict_to_url(param: dict):
    # Encode the dictionary with url-safe-base64.
    str_param = json.dumps(param)
    url_param = base64.urlsafe_b64encode(bytes(str_param, encoding="utf8"))
    return bytes.decode(url_param, encoding="utf8")


def url_to_dict(url_param: str):
    # Decode url-safe-base64 encoded string.
    bytes_param = bytes(url_param, encoding="utf8")
    str_param = bytes.decode(base64.urlsafe_b64decode(bytes_param), encoding="utf8")
    return json.loads(str_param)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorBinOp
from .utils import arithmetic_operand


@arithmetic_operand
class TensorFloorDiv(TensorBinOp):
    _op_type_ = OperandDef.FLOORDIV
    _func_name = "floor_divide"

    @classmethod
    def _is_sparse(cls, x1, x2):
        if hasattr(x1, "issparse") and x1.issparse():
            if x2 != 0:
                return True
            else:
                raise ZeroDivisionError("float division by zero")
        return False


@infer_dtype(np.floor_divide)
def floordiv(x1, x2, out=None, where=None, **kwargs):
    """
    Return the largest integer smaller or equal to the division of the inputs.
    It is equivalent to the Python ``//`` operator and pairs with the
    Python ``%`` (`remainder`), function so that ``b = a % b + b * (a // b)``
    up to roundoff.

    Parameters
    ----------
    x1 : array_like
        Numerator.
    x2 : array_like
        Denominator.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated array is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor
        y = floor(`x1`/`x2`)


    See Also
    --------
    remainder : Remainder complementary to floor_divide.
    divmod : Simultaneous floor division and remainder.
    divide : Standard division.
    floor : Round a number to the nearest integer toward minus infinity.
    ceil : Round a number to the nearest integer toward infinity.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.floor_divide(7,3).execute()
    2
    >>> mt.floor_divide([1., 2., 3., 4.], 2.5).execute()
    array([ 0.,  0.,  1.,  1.])
    """
    op = TensorFloorDiv(**kwargs)
    return op(x1, x2, out=out, where=where)


@infer_dtype(np.floor_divide, reverse=True)
def rfloordiv(x1, x2, **kwargs):
    op = TensorFloorDiv(**kwargs)
    return op.rcall(x1, x2)
